/*
	SPP: definition and expansion routines.
	
	source:  SPPdef.c
	started: October 22, 1985
	version: March 16, 1996.
*/

#include "SPP.h"

#include "SPParg.h"
#include "SPPdef.h"
#include "SPPmain.h"
#include "SPPmst.h"
#include "SPPnext.h"
#include "SPPpr.h"
#include "SPPres.h"
#include "SPPsem.h"
#include "SPPtok.h"

#include "SPPerr.h"	/* Library additions. */
#include "SPPobj.h"
#include "SPPstr.h"

/*
	The following are used for communication with io_next.
*/

bool	m_flag = FALSE;			/* TRUE if expanding a macro.	*/
char	macro_buf[MAX_RTEXT];	/* Final macro buffer.		*/
char *	p_rescan;				/* Pointer into rescan_buf[].	*/

/*
	Forward declarations of internal routines.
*/

static void	aexpand		(char *name, int *nargs, char **p1, char *out,
							char **p2);
static void	aparse		(char *name, int *nargs, char *out, char **p);
static bool	a1parse		(char *name, char *out, int max_out);
static int	arg_search	(char *name, int nargs, char ** argps);
static void	disable_name	(char *name);
static void	enable_name	(void);
static bool	is_disabled	(char *name);
static void	expand		(char *name, int nargs, char *rtext, char *out,
							int max_out);
static void	rescan		(char *name, char *in, char *out, int max_out,
							bool bounded_flag);
static void	substitute	(char *name, int nargs, char *in, char **args,
							char *out, int max_out);

/*
	Define static variables used only by this module.
*/
#define MAX_MDEPTH		100		/* Max depth of macro expansions.	*/

static char *	x_stack [MAX_MDEPTH];	/* Pointers to disabled macro names. */
static int		x_count = 0;			/* Number of disabled macro names.   */

/*
	Externally visible routines.
*/


/*
	
	Initialize this module.
*/
void
def_init_again(void)
{
	m_flag = FALSE;		/* Not expanding a macro.		*/
	p_rescan = NULL;	/* Pointer into rescan_buf[].	*/
	x_count = 0;		/* No disabled macros.			*/

}

/*
	Check an identifier seen at the outer level to see whether it
	could be a macro call and expand it and return TRUE if so.
*/
bool
outer_expand(char *name, bool old_mflag)
{
	FTAG("outer_expand");

	char 	*p1, *p2;
	int		i;
	char *	rtext;					/* Replacement text for name.	*/
	int		nargs;					/* Required number of args.		*/
	char	temp_buf[MAX_RTEXT];	/* Temp spot for macro.			*/	
	
	TRACEPB(ftag, es(name); enl());
	
	if (m_flag || old_mflag || !mst_lookup(name, &rtext, &nargs)) {
		STATX(ftag);
		return (FALSE);
	}
	else if (nargs != -1) {
		skip_ws(TRUE);
		if (ch != '(') {
			warning(
				es("Function-like macro "); es(name);
				es(" appears without arguments"));

			STATX(ftag);
			return FALSE;
		}
	}

	/* Disable Sherlock output when a Sherlock macro is seen. */
	if (is_sherlock(name)) {
		so_disable();
	}

	/*
		Expand the macro into temp_buf[].
		This will use ONLY file characters.
	*/
	expand(name, nargs, rtext, &temp_buf[0], MAX_RTEXT);
	
	if (*rtext >= MAX_RTEXT) {
		fatal(es("outer_expand: expand overflow"));
	}

	/*
		Expand will have changed ch to point past any args.
		Use sys_fpb so that
		1.  The old ch will appear just AFTER the macro expansion.
		2.  It will not be mistaken added to the rescan buffer.
	*/
	io_fpb(ch);

	/*
		Fully rescan temp_buf[] into macro_buf[].
		Use an unbounded rescan, i.e., allow file characters to
		complete a partial list of actual parameters which may
		be started in the rescanned text.
	*/
	p_rescan = NULL;
	m_flag   = FALSE;
	rescan(name, &temp_buf[0], &macro_buf[0], MAX_RTEXT, FALSE);

	/* Make sure we save the now current ch. */
	io_fpb(ch);

	/* Delete any internal flags from the output. */
	TRACEP(ftag,
		es("macro1: "); es_str(&macro_buf[0]); enl());

	p1 = p2 = &macro_buf[0];
	i = 0;
	while (*p2) {
		i++;
		if (i >= MAX_RTEXT) {
			fatal(es("outer_expand: rescan overflow"));
		}
		if (*p2 == EXPAND_OFF || *p2 == EORT) {
			p2++;
		}
		else {
			*p1++ = *p2++;
		}
	}
	*p1 = '\0';
	TRACEP(ftag, es("macro2: "); es_str(&macro_buf[0]); enl());

	/*
		Inhibit any further macro expansion until the end of the
		current local input stream.
	*/

	if (m_flag) {
		fatal(es("outer_expand: Can't happen"));
	}

	/*
		Indicate that the macro buffer has a macro.
	*/
	m_flag = TRUE;
	p_rescan = &macro_buf[0];

	/* Reload ch from the global macro buffer. */
	io_next();

	STATX(ftag)
	return TRUE;
}

/*
	Handle the #define directive by parsing the statement, and entering
	the macro's name, number of arguments and replacement text into the
	macro table.

	Formal arguments are found and replaced by a flag byte followed by the
	number of the formal argument.
*/
void
pp_def(void)
{
	FTAG("pp_def");
	
	int	i;

	char	name[MAX_SYMBOL];	/* Name of macro.				*/

	char	rtext [MAX_RTEXT];	/* Replacement text.			*/
	int		rcount = 0;			/* Number of chars in rtext[].	*/

	char	atext [MAX_ATEXT];	/* Text of formal parameters.	*/
	int		acount = 0;			/* Number of chars in atext[].	*/

	char *	argps [MAX_NARGS];	/* Pointers into atext[].	 	*/
	int	nargs = 0;				/* Number of entries in argps[]. */

	TICKB(ftag);

	/* Make sure the name is present. */
	if (!isid1(ch)) {
		error(es("#define ignored--no symbol name given"));
		skip_1line();
		return;
	}

	/* Get the macro name into name. */
	t_id(&name[0], MAX_SYMBOL);
	TRACEPN("pp_def", es("macro name is "); es(name); enl());
	
	if (ch != '(') {
		/* Indicate no argument list and continue. */
		nargs = -1;
		goto gettext;
	}
	else {
		io_next();
	}

	if (ch == ')') {
		io_next();
		nargs = 0;
		goto gettext;
	}

	/* 
		Put the formal arguments into dynamic storage.
		Set pointers to the arguments argps[].
	*/
	while (nargs < MAX_NARGS) {

		skip_ws(TRUE);

		if (!isid1(ch)) {
			error(es("#define ignored--formal arg must be an identifier"));
			skip_1line();
			
			STATX(ftag);
			return;
		}

		/* Copy one formal arg to atext[]. */
		t_id(&t_symbol[0], MAX_SYMBOL);
		argps [nargs++] = &atext[acount];
		strcpy(atext + acount, t_symbol);
		acount += strlen(t_symbol)+1;

		TRACEPN("pp_def",
			es("formal ["); eint(nargs-1);
			es("] is "); eptr(argps [nargs-1]));

		/* Bug fix: 11/3/88 */
		skip_ws(TRUE);
			
		if (ch == ')') {
			io_next();
			goto gettext;
		}
		else if (ch == ',') {
			io_next();
		}
	}
	error(es("#define ignored--too many arguments"));
	skip_1line();

	STATX(ftag);
	return;


	/*
		At this point, nargs contains the number of formal arguments,
		or -1 if no argument list was given.
		0 is allowed and means that the argument list was ().
	*/

gettext:

	TRACEPN("pp_def", es("nargs is "); eint(nargs); enl());

	skip_bl();

	/*
		Put the replacement text into rtext[].
		The text is tokenized, which means that partial comments
		or strings are not valid.

		Replace formal arg n by n FLAG on the fly.
	*/

	for (rcount = 0;;) {
		switch (ch) {

		case END_FILE:
		case '\n':
			goto done;

		/* Replace a sequence of white space by a blank. */
		case ' ':
		case '\t':
			io_next();
			if (rcount > 0 && rtext[rcount-1] != ' ') {
				rtext[rcount++] = ' ';
			}
			continue;

#ifdef BACKSLASH_NEWLINE
		/* Look for continuation of definitions. */
		case '\\':
			io_next();
			if (ch == '\n') {
				io_next();
				if (rcount > 0 && rtext[rcount-1] != ' ') {
					rtext[rcount++] = ' ';
				}
				bump_line();
			}
			else {
				warning(es("Backslash not followed by newline"); enl());
				rtext[rcount++] = '\\';
			}
			continue;
#endif

		/* Eliminate standard and single-line comments. */
		case '/':
			io_next();
			if (ch == '*') {
				/* Skip a regular comment. */
				if (rcount > 0 && rtext[rcount-1] != ' ') {
					rtext[rcount++] = ' ';
				}
				io_next();
				skip_comment();
			}
			else if (slc_flag && ch == '/') {
				/* 8/3/89 Skip a single-line comment. */
				if (rcount > 0 && rtext[rcount-1] != ' ') {
					rtext[rcount++] = ' ';
				}
				while (ch != END_FILE && ch != '\n') {
					io_next();
				}
				goto done;
			}
			else {
				rtext[rcount++] = '/';
			}
			continue;

		/* Handle complete strings. */
		case '"':
		case '\'':
			t_string(&rtext[rcount], MAX_RTEXT-rcount, TRUE);
			rcount += t_length;
			continue;

		/* Handle # formal param and ## */
		case '#':
			io_next();
			if (ch == '#') {
				io_next();
				rtext[rcount++] = CONCAT_FLAG;
				continue;
			}

			/* We expect an identifier here. */
			/* bug fix: 2/15/89 */
			if (ch == ' ' || ch == '\t') {
				rtext[rcount++] = ' ';
				skip_bl();
			}
			if (isid1(ch)) {
				t_id(&rtext[rcount], MAX_RTEXT-rcount);
				i = arg_search( rtext + rcount, nargs,
						&argps[0]);
				if (i >= 0) {
					/* Replace id with flags. */
					rtext[rcount++] = POUND_FLAG;
					rtext[rcount++] = i + ARG_OFFSET;
				}
				else {
					/* Accept id. */
					rcount += t_length;
				}
			}
			else {
				/* bug fix: 2/15/89 */
				warning(es("Id expected after '#'"); enl());
			}
			continue;

		default:

			/* Possible formal parameter. */
			if (isid1(ch)) {
				t_id(&rtext[rcount], MAX_RTEXT-rcount);
				i = arg_search(	rtext + rcount, nargs,
						&argps[0]);
				if (i >= 0) {
					rtext[rcount++] = ARG_FLAG;
					rtext[rcount++] = i+ARG_OFFSET;
				}
				else {
					rcount += t_length;
				}
			}
			else {
				rtext[rcount++] = ch;
				io_next();
			}
			continue;
		}
	}

done:

	/* This check is made only here to save a little time. */
	if (rcount >= MAX_RTEXT) {
		fatal(es("Replacement text of macro is too long"));
	}

	/* Strip at most one blank off the end of the definition. */
	if (rtext[rcount - 1] == ' ') {
		rcount--;
	}

	/* Terminate the replacement text properly. */
	rtext[rcount] = '\0';

	TRACEPN("pp_def", es("rtext: <"); es_str(rtext); es(">"); enl());

	/*
		Enter the symbol, replacement text and number of arguments
		into the macro table.

		Note that mst_enter allocates space for copies of name and
		rtext so the fact that they are on the stack doesn't matter.
	*/
	mst_enter(&name[0], rtext, nargs);
	skip_pp();

	STATX(ftag);
}


/*
	Internal Routines.
*/

/*
	Macro expand all arguments in old_argps and leave them in new_atext[].
	Put pointers to them in new_argps[].
*/
static void
aexpand(
	char *	name,			/* Macro name whose args are being expanded.	*/
	int *	nargs,			/* Number of args the macro should have.		*/
	char *	old_argps[],	/* Pointers to unexpanded tokens for each arg.	*/
	char 	new_atext[],	/* Storage for expanded args.					*/
	char *	new_argps[])	/* Pointers to expanded args.					*/
{
	FTAG("aexpand");

	char	save_ch;
	int		i;
	char *	new;
	int		new_count = 0;

	TRACEPB(ftag,
		es(name); ecs();
		eint(*nargs); ecs();
		eptr(old_argps); ecs();
		eptr(new_atext); ecs();
		eptr(new_argps); enl());

	/*
		Save the current value of ch!
		rescan will change ch, and it must be restored afterwards.
	*/
	save_ch = ch;

	/* Rescan all arguments. */
	for (i = 0; i < *nargs; i++) {
		new = &new_atext[new_count];
		new_argps[i] = new;

		/* Do a bounded rescan. */
		rescan(	"<NO_NAME>", old_argps[i], new,
			MAX_RTEXT-new_count, TRUE);

		new_count += strlen(new)+1;
	}

	/* Restore ch */
	ch = save_ch;

	STATX(ftag);
}

/*
	Parse a list of actual arguments into atext[] and put pointers to
	the arguments in argps[].  
	Set *nargs to -1 on error.
	The size of atext[] is MAX_ATEXT and the size of argps is MAX_ARGPS.
*/
static long call_start;		/* Line where arguments started.	*/

static void
aparse(
	char *	name,		/* Macro whose arguments are being parsed.		*/
	int *	nargs,		/* Pointer to arg count.						*/
	char 	atext[],	/* Buffer where UNEXPANDED arguments are put.	*/
	char *	argps[])	/* Array of pointers to arguments.				*/
{
	FTAG("aparse");

	int		arg_c	= 0;	/* Number of arguments parsed so far.	*/
	int		length	= 0;	/* Total length of arguemnts parsed.	*/
	bool	flag;

	if (*nargs == -1) {
		return;
	}

	TRACEPB(ftag,
		es(name); ecs();
		eint(*nargs); ecs();
		eptr(atext);  ecs();
		eptr(argps); enl());

	if (ch != '(') {
		fatal(es("aparse: Can't happen"));
	}
	else {
		io_next();
	}

	/* Save starting line number of the macro call. */
	call_start = err_line;

	if (ch == ')') {
		io_next();
		goto check;
	}

	for(;;) {
		/* Set pointer to the start of the argument. */
		argps[arg_c] = &atext[length];

		/* Parse the argument. */
		flag = a1parse(name, &atext[length], MAX_ATEXT-length);
		if (flag || ch == END_FILE) {
			*nargs = -1;

			STATX(ftag);
			return;
		}
		else {
			length += strlen(argps[arg_c])+1;
			arg_c++;
		}

		if (ch == ')') {
			io_next();
			break;
		}
		else if (ch == ',') {
			io_next();
		}
		else {
			/* Error detected in a1parse. */
			break;
		}
	}

check:
	if (arg_c != *nargs) {
		error(
			es("Call to macro "); es(name);
			if (call_start != err_line) {
				es(" starting at line "); elong(call_start);
			}
			es(" requires "); eint(*nargs); es(" arguments"));

		*nargs = -1;
	}

	TRACEPX(ftag,
		{int i;
		if (*nargs != 0) enl();
		for(i = 0; i < *nargs; i++) {
			es("arg["); eint(i); es("]: @"); eptr(&argps[i]);
			es("="); eptr(argps[i]); es(": <"); es_str(argps[i]); es(">\n");
		}});
}

/*
	Parse one actual argument into out[].
	The size of out is max_length.
	Return TRUE if an error was seen.

	Formal arguments are simply identifiers, which are handled by t_id(),
	but actual arguments are LISTS of tokens separated by commas.  As an
	added twist, commas inside single or double quotes, or commas which are
	"protected" by additional parentheses do NOT separate actual args.
	Thus, each of the following calls have to M have ONE actual argument:

		M(a)
		M(a * N(c,b))
		M(',')
		M("a,b")
		M((a,b))
		M((a,")",b))

	This routine changes all comments and white space to a single blank.
*/
static bool
a1parse(
	char *	name,			/* Name of the macro being expanded.	*/
	char *	out,			/* Output buffer.			*/
	int		max_out)		/* Size of out[].			*/
{
	FTAG("a1parse");

 	int count;			/* Number of characters in out[].	*/
	int plevel;			/* Parenthesis level.				*/

	TRACEPB(ftag,
		es(name); ecs(); eptr(out); ecs(); eint(max_out); enl());

	/* No parens have been seen yet. */
	plevel = 0;

	for(count = 0;;) {

		/* Make sure there is room for one more. */
		if (count >= max_out) {
			goto toolong;
		}

		switch (ch) {

		case END_FILE:
			goto runon;

		case '\n':
			io_next();
			bump_line();
			if (count > 0 && out[count-1] != ' ') {
				out[count++] = ' ';
			}
			continue;

		/* Convert a sequence of white space to one blank. */
		case ' ':
		case '\t':
			io_next();
			if (count > 0 && out[count-1] != ' ') {
				out[count++] = ' ';
			}
			continue;

#ifdef BACKSLASH_NEWLINE
		case '\\':
			io_next();
			if (ch == '\n') {
				io_next();
				if (count > 0 && out[count-1] != ' ') {
					out[count++] = ' ';
				}
				bump_line();
			}
			else {
				out[count++] = '\\';
			}
			continue;
#endif

		case ',':
			if (plevel == 0) {
				goto end_arg;
			}
			else {
				out[count++] = ch;
				io_next();
			}
			continue;
				
		case ')':

			if (plevel == 0) {
				goto end_arg;
			}
			else {
				plevel--;
				out[count++] = ch;
				io_next();
				continue;
			}

		case '(':
			plevel++;
			out[count++] = ch;
			io_next();
			continue;

		case '"':
		case '\'':
			t_string(&out[count], max_out-count, TRUE);
			count += t_length;
			if (count >= max_out) {
				goto toolong;
			}
			continue;

		case '/':
			io_next();
			if (ch == '*') {
				io_next();
				skip_comment();

				/* Change a comment into one blank. */
				if (count > 0 && out[count] != ' ') {
					out[count++] = ' ';
				}
			}
			else {
				out[count++] = '/';
			}
			continue;

		default:
			if (isid1(ch)) {
				t_id(&out[count], max_out-count);
				count += t_length;
				if (count >= max_out) {
					goto toolong;
				}
			}
			else {
				out[count++] = ch;
				io_next();
			}
		}
	}

end_arg:
	/* Finish off the argument. */
	out[count] = '\0';
	TRACEP(ftag, es_str(out); ecnl());
	
	STATX(ftag);
	return FALSE;

runon:
	error(
		es("Runon macro call");
		if (call_start != err_line) {
			es(" at line "); elong(call_start);
		}
		es("--last arg set to null"));
	
	STATX(ftag);
	return TRUE;

toolong:
	error(
		es("Macro argument starting at line ");
		elong(call_start); es(" is too long"));
	
	STATX(ftag);
	return TRUE;
}

/*
	Search an array of formal arguments for a match.
*/
static int
arg_search(
	char *	name,	/* Name to search for.	*/
	int		nargs,	/* Number of args.		*/
	char **	argps)	/* Array of args.		*/
{
	int i;

	/* See if the id is a formal arg. */
	for (i = 0; i < nargs ; i++) {
		if (str_eq(name, argps[i])) {
			return i;
		}
	}
	return -1;
}

/*
	Disable the expansion of one name.
*/
static void
disable_name(char * name)
{
	TRACEP("disable_name", es(name); enl());

	if (x_count < MAX_MDEPTH) {
		x_stack [x_count++] = name;
	}
}

/*
	Enable the expansion of the last disabled name.
*/
static void
enable_name(void)
{
	TICK("enable_name");

	if (x_count == 0) {
		fatal(es("enable_name: Can't happen"));
	}
	x_count--;
}

/*
	Fully expand the macro described by name, nargs, rtext.
*/
static void
expand(
	char *	name,		/* Name of macro expanded.		*/
	int		nargs,		/* Required number of args.		*/
	char *	rtext,		/* Replacement text for name.	*/
	char *	out,		/* Output buffer.				*/
	int		max_out)	/* Size of out[].				*/
{
	FTAG("expand");

	char *	old_atext;	/* Buffer for actual arguments.	   */
	char *	new_atext;	/* Buffer for expansion of args.   */	

	char *	old_argps[MAX_NARGS];	/* Pointers into old_atext[].	   */
	char *	new_argps[MAX_NARGS];	/* Pointers into new_atext[].	   */

	TRACEPB(ftag,
		es(name); ecs(); eint(nargs);
		es(", rtext: <"); es_str(rtext); es(">, ");
		eptr(out); ecs(); eint(max_out);
		es(" ch: "); es_ch(ch); enl());

	/*
		6/27/89
		Allocate memory here so we don't crash the stack.
	*/
	old_atext = obj_alloc(MAX_ATEXT,
		OBJ_EXPAND_OLD_ATEXT+M_PROT+M_NOFILL);
		
	new_atext = obj_alloc(MAX_ATEXT,
		OBJ_EXPAND_NEW_ATEXT+M_PROT+M_NOFILL);

	/* Parse all arguments to the macro. */
	aparse(name, &nargs, old_atext, old_argps);

	/* Fully macro expand all arguments. */
	aexpand(name, &nargs, &old_argps[0], &new_atext[0], &new_argps[0]);

	/* Substitute all expanded arguments into rtext[], giving out[]. */
	substitute(name, nargs, rtext, &new_argps[0], &out[0], max_out);

	/* Free all memory. */
	obj_free(new_atext);
	obj_free(old_atext);

	STATX(ftag);
}

/*
	Return TRUE if macro expansion of name has been disabled.
*/
static bool
is_disabled(char * name)
{
	FTAG("-is_disabled_v");

	int i;

	TRACEPB(ftag, 
		es(name); es(", x_count: "); eint(x_count); enl());

	for (i = x_count-1; i >= 0; i--) {
		TRACEP(ftag, 
			es("compare: <"); es(name); es("> <"); es(x_stack[i]); enl());

		if (str_eq(name, x_stack[i])) {
			STATX(ftag);
			return TRUE;
		}
	}
	
	STATX(ftag);
	return FALSE;
}

/*
	This is it: the guts of macro expansion.
	Fully rescan in[] to out[], looking for macro calls.

	A key distinction:  bounded versus unbounded rescanning.

	The CALLER of rescan must save and restore ch as appropriate.
	(Note that rescan is called recursively from within, so it must
	save ch in that case just as any other caller must.

	Note that rescan will completely process rescan buffer in all cases.
	For unbounded rescans, the previous rescan buffer is appended to in[].
	For bounded rescans, the previous rescan buffer is kept separate.
*/
static void
rescan(
	char *	name,			/* Name of macro being rescanned.	*/
	char *	in,				/* Text to be rescanned.			*/
	char *	out,			/* Output buffer.					*/
	int		max_out,		/* Size of out[].					*/
	bool	bounded_flag)	/* TRUE if bounded rescan.			*/
{
	FTAG("rescan");
	FTAGV("-rescan_v");

	int		save_mflag;			/* Saved m_flag.	*/
	char *	save_prescan;		/* Saved p_rescan.	*/

	char *	local_in;	/* Local rescan buffer.	*/
	char *	expand_buf;	/* Expansion buffer.	*/
	char *	id_buf;		/* Saved rescan id.		*/

	int		out_len = 0;	/* Index into out[].		*/
	char *	rtext;			/* Ptr to replacement text.	*/
	int		nargs;			/* Required number of args.	*/
	int		i, n;

	TRACEPB(ftag, 
		es(name); ecs();
		es_str(in); ecs();
		eptr(out); ecs();
		eint(max_out); ecs();
		ebool(bounded_flag);
		es(" ch: "); es_ch(ch); enl());

	/*
		6/27/89
		Allocate buffers here so we don't crash the stack.
	*/
	local_in   = obj_alloc(MAX_RTEXT, OBJ_RESCAN_LOCAL_IN+M_PROT+M_NOFILL);
	expand_buf = obj_alloc(MAX_RTEXT, OBJ_RESCAN_EXPAND_BUF+M_PROT+M_NOFILL);
	id_buf     = obj_alloc(MAX_SYMBOL, OBJ_RESCAN_ID_BUF+M_PROT+M_NOFILL);
	
	/*
			Check size of input.
	*/
	n = strlen(in);
	if (max_out < 20 || n + 2 > MAX_RTEXT || n + 2 > max_out) {
		error(es("Macro expansion too big--set to empty string"));
		
		goto done;
	}
	
	TRACEN("no_rescan",
		out[0] = EORT;
		out[1] = '\0';
		goto done;
	);

	if (bounded_flag) {
		/* Save old rescan buffer and use in[] as the new buffer. */
		save_mflag   = m_flag;
		save_prescan = p_rescan;
			
		/*
			Indicate the end of the buffer with EORT flag.
			Copy in[] to local_in[] so we have room for it.
		*/
		p_rescan = &local_in[0];
		strcpy(&local_in[0], in);
		i = strlen(&local_in[0]);
		local_in[i++] = EORT;
		local_in[i]   = '\0';
	}
	else {
		/*
			The key programming trick in this whole file.
			Set rescan to old rescan appended behind in[].
		*/
		if (!m_flag) {
			p_rescan = in;
		}
		else {
			strcpy(local_in, in);
			strcat(&local_in[0], p_rescan);
			p_rescan = &local_in[0];
		}
	}

	/* Get input from rescan buffer. */
	m_flag = TRUE;

	TRACEP(ftagv, es("p_rescan: <"); es_str(p_rescan); es(">\n"));

	/*
		Note1:	At this point, io_next() will return the characters
		that have just been placed in the rescan buffer pointed to by
		p_rescan.  This means that t_id(), t_string(), etc.  may be
		used to parse the rescan buffer.

		Note2:  For bounded rescans, the rescan buffer ends with EORT,
		so io_next() will return EORT when the rescan buffer is
		exhausted.  This means that t_string() and skip_comment() will
		complain about "unexpected end of input," which is as it
		should be.
	*/

	/* Disable further expansions of name for the duration. */
	disable_name(name);

	/*
		Scan the buffer until it is exhausted or a macro call is seen.
		(The recursive call to rescan() will finish off the buffer.)
	*/
	out_len = 0;
	io_next();
	while (m_flag) {
	
		obj_checkObj(local_in, "rescan: local_in");
		obj_checkObj(expand_buf, "rescan: expand_buf");
		obj_checkObj(id_buf, "rescan: id_buf");
		
		if (ch == EORT) {
			break;
		}
		else if (ch == EXPAND_OFF) {
			/* Inhibited identifier. */
			out[out_len++] = ch;
			io_next();
			t_id(&out[out_len], max_out-out_len);
			out_len += strlen(&out[out_len]);
		}
		else if (ch == '"' || ch == '\'') {
			/* Handle complete strings. */
			t_string(&out[out_len], max_out-out_len, TRUE);
			out_len += t_length;
		}
		else if (isid1(ch)) {
			/* Handle identifiers. */
			t_id(&id_buf[0], MAX_SYMBOL);

			/*
				Possible inner macro call.
				Do not adjust out_len until we know for sure.
			*/
			if (!mst_lookup(&id_buf[0], &rtext, &nargs)) {
				/* Not a macro call. */
				strcpy(&out[out_len], &id_buf[0]);
				out_len += t_length;
				continue;
			}
			else if (is_disabled(&id_buf[0])) {
				/*
					Disabled macro.
					Inhibit ALL further expansion.
				*/
				out[out_len++] = EXPAND_OFF;
				strcpy(&out[out_len], &id_buf[0]);
				out_len += t_length;
				continue;
			}
			else if (nargs > -1) {
				skip_ws(TRUE);
				/*
					Function-like macro name without args
					are ok in bounded expansions.
				*/
				if (ch != '(') {
					if (ch != EORT && !bounded_flag) {
						warning(
							es("Function-like macro "); es(&id_buf[0]);
							es(" appears without arguments"));
					}
					strcpy(&out[out_len], &id_buf[0]);
					out_len += t_length;
					continue;
				}
			}

			/* Valid macro call. */

			out[out_len] = '\0';

			/* Expand the text into expand_buf[]. */
			expand(	&id_buf[0], nargs, rtext,
				&expand_buf[0], MAX_RTEXT);

			if (m_flag) {
				/* Append current ch to expand_buf[]. */
				i = strlen(&expand_buf[0]);
				expand_buf[i++] = ch;
				expand_buf[i] = '\0';
			}
			else {
				/*
					Don't needlessly suck file characters
					into the expansion.
				*/
				io_fpb(ch);
			}

			/* Rescan will append the rest of p_rescan. */
			TRACEP(ftag, es("recursive call. ch: "); es_ch(ch); enl());

			rescan(	&id_buf[0], &expand_buf[0],
				&out[out_len], MAX_RTEXT-out_len, FALSE);
				
			obj_checkObj(expand_buf, "rescan2: expand_buf");

			out_len=strlen(&out[0]);

			/* Use the ch that the recursive call left behind. */
			continue;
		}
		/* Comments might appear here due to concatenation. */
		else if (ch == '/') {
			io_next();
			if (ch == '*') {
				io_next();
				skip_comment();
			}
			else {
				out[out_len++] = '/';
			}
		}
		else {
			out[out_len++] = ch;
			io_next();
		}
	}

	out[out_len] = '\0';
	
	if (out_len >= max_out) {
		fatal(es("rescan: out buffer overflow"));
	}
	
	TRACEP(ftag, es("out: <"); es_str(out); es(">\n"));

	enable_name();

	/* Restore the b_rescan global. */
	if (bounded_flag) {
		/* Restore the old rescan buffer. */
		m_flag   = save_mflag;
		p_rescan = save_prescan;
	}
	else {
		/* Everything is complete. */
		m_flag = FALSE;
	}

	/* Free buffers. */
done:
	obj_free(id_buf);
	obj_free(expand_buf);
	obj_free(local_in);

	STATX("rescan");
}

/*
	Substitute expanded arguments from argps[] into in[] leaving out[].
	6/27/89 bug fix: a check is made to see if out is large enough.
*/
static void
substitute(
	char *	name,		/* The name of the macro being expanded.	*/
	int		nargs,		/* Number of args for the macro.			*/
	char *	rtext,		/* Replacement text of the macro.			*/
	char **	argps,		/* Pointer to fully expanded arguments.		*/
	char *	out,		/* Output buffer.							*/
	int		max_out) 	/* Size of output buffer.					*/
{
	FTAG("substitute");

	int		limit;
	int		c, i, count;
	int		arg_num;
	char	*p, *p2;
	char	str_buf[MAX_SYMBOL];

	TRACEPB(ftag,
		es(name); ecs(); eint(nargs);
		es(", <"); es_str(rtext); es(">, ");
		eptr(argps); ecs(); eptr(out);
		es(", max_out: "); eint(max_out); enl());

	/* Make SURE the output buffer is never over-run. */
	max_out -= 4;
	if (max_out < 10) {
		goto run_on;
	}

	/*
		Put the final replacement text into out[].
		Replace ARG_FLAG n by arg n.
		Replace POUND_FLAG n by the stringized arg n.
		Delete CONCAT_FLAG and surrounding whitespace.
	*/
	limit = strlen(rtext);
	count = 0;
	for (i = 0; i < limit;) {

		/*
			Check for run-on expansion.
			This is actually possible.
		*/
		if (count >= max_out) {
			goto run_on;
		}

		c = rtext[i++];
		if (c == ARG_FLAG) {

			/* Copy the argument. */
			arg_num = rtext[i++] - ARG_OFFSET;
			if (arg_num >= nargs) {
				/* Previous user error. */
				out[count++] = ' ';
				continue;
			}

			/* Bug fix: 6/27/89 */
			if (count + strlen(argps[arg_num]) >= max_out) {
				goto run_on;
			}
			strcpy(out + count, argps [arg_num]);
			count += strlen(argps[arg_num]);
		}
		else if (c == POUND_FLAG) {

			/* Stringize an actual argument */
			arg_num = rtext[i++] - ARG_OFFSET;
			if (arg_num >= nargs) {
				/* Previous user error. */
				out[count++] = ' ';
				continue;
			}
			out[count++] = '"';
			for(p = argps[arg_num]; *p; ) {
				/* bug fix: 2/15/89 */
				if (*p == '"' || *p == '\'') {
					/* Copy the string. */
					p += in_string(p, &str_buf[0], MAX_SYMBOL);
	
					/* Stingize the string. */
					for (p2 = &str_buf[0]; *p2; ) {
						/* 8/1/89 better here. */
						if (count >= max_out-1) {
							goto run_on;
						}
						if (*p2 == '\\' || *p2 == '"') {
							out[count++] = '\\';
							out[count++] = *p2++;
						}
						else {
							out[count++] = *p2++;
						}
					}
				}
				else {
					out[count++] = *p++;
				}
			} /* end for */
			out[count++] = '"';
		}
		else if (c == CONCAT_FLAG) {

			/* Delete preceding whitespace. */
			while (count-1 >= 0 && out[count-1] == ' ') {
				count--;
			}

			/* Delete following whitespace. */
			while (i < limit && rtext[i] == ' ') {
				i++;
			}
		}
		else {
			out[count++] = c;
		}
	}

	/* Append current ch to output. */
	out[count] = '\0';

	TRACEPX(ftag, es("out: <"); es_str(out); es(">\n"));
	return;

run_on:
	error(
		es("Macro expansion for "); es(name);
		es(" is too long--set to empty string"));

	/* Truncate the ENTIRE macro. */
	out[0] = '\0';
	STATX("substitute");
}
