/*
	SPP: macro symbol table routines.

	source:  SPPmst.c
	started: September 22, 1985
	version: March 16, 1996.
*/

#include "spp.h"

#include "SPPmst.h"
#include "SPPpr.h"

#include "SPPenv.h"		/* Library additions. */
#include "SPPerr.h"
#include "SPPlst.h"
#include "SPPobj.h"
#include "SPPstr.h"
#include "SPPtime.h"

/*
	Define the hash table used to access macro table nodes.
*/
#define MAC_PRIME 101

typedef struct mst_struct mst_node;

static struct mst_struct {
	LIST;
	int	 		mst_nargs;	/* Number of args in text.		*/
	char *		mst_name;	/* Pointer to name.				*/
	char *	 	mst_text;	/* Pointer to replacement text. */
};

static mst_node * mst_ht [MAC_PRIME];
static mst_node * def_save;


/*
	Define the "initial undef list" for
	arguments undefined on the command line.
*/
typedef struct u_node unode;
static struct u_node {
	LIST;
	bool		u_flag;		/* TRUE: not seen yet.	*/
	char *		u_name;		/* Name of undef.		*/
};

static unode * undef_list;	/* List header for current undef list.	*/
static unode * undef_save;	/* List header for saved undef list.	*/

/*
	Declare local routines.
*/
static int	mst_hash	(char *);
static void	mst_udump	(lst_node * lnp);
static void mst_ndump	(lst_node * lnp);

/*
	Remove ALL instances of a symbol from the symbol table.
	This ensures that an #undef will expose any non-macro version
	of a routine.
*/
#undef symbol	/* Symbol is defined in the Mac world to be a font. */

void
mst_delete (reg char * symbol)
{
	FTAG("mst_delete");

	reg mst_node *	bp;
	reg mst_node *	bp2;
	reg int			hash;

	TRACEP(ftag, es("entry: symbol=<"); es(symbol); es(">\n"));
	
	hash = mst_hash(symbol);
	lst_safe_every(mst_node, bp, bp2, &mst_ht[hash]) {
		if(str_eq(symbol, bp -> mst_name)) {

			TRACEP(ftag, es("deletes "); es(symbol); es("\n"));
			
			lst_remove(bp, &mst_ht[hash]);
			obj_free(bp);
		}
	}
}

/*
	Place a macro in the symbol table along with the number of arguments
	to the macro and a pointer to the macro's replacement text.

	Multiple entries for a symbol can exist in which case the last entry
	defined is active.

	This code conforms to the new standard.  It rejects duplicate 
	definitions of a macro unless the new definition is exactly the
	same as the old.  

	This code also checks the "initial undef list" to see if the first
	definition of a macro should be ignored.  If the macro name is
	found on the list, it is removed from the list and no entry is made
	in the macro symbol table.
*/
void
mst_enter (
	reg char *	symbol,
	reg char *	txt,
	reg int		nargs)
{
	FTAG("mst_enter");

	reg mst_node *	bp;
	reg unode *		up;
	reg int 		hash;

	TRACEP(ftag,
		es(symbol); es(", <"); es_str(txt); es(">, ");
		eint(nargs); enl());

	/*
		Search the global "initial undef list" (created by command line
		-u arguments) to see if the initial definition of name should
		be rejected.  If the name appears on the list, remove it from
		the list and do not enter the name in the macro table.
	*/
	
	lst_every(unode, up, &undef_list) {
		if (str_eq(symbol, up -> u_name) && up -> u_flag) {
		
			/* Just clear the flag. */
			up -> u_flag = FALSE;
			return;
		}
	}

	/* Search down the list of mst_nodes. */
	hash = mst_hash(symbol);

	lst_every(mst_node, bp, &mst_ht[hash]) {
	
		/* Reject redefinitions of a macro unless identical. */
		if (str_eq (symbol, bp -> mst_name)) {
			if (nargs == bp -> mst_nargs) {
				if (str_eq(txt, bp -> mst_text)) {
					return;
				}
			}
			error(
				es("Non-identical redefinition of ");
				es(symbol); es(" ignored"); enl());

			return;
		}
	}

	/* Dynamically allocate space for node. */
	bp = obj_new(mst_node, OBJ_MST_STRUCT+M_NODUMP);
	
	/* Fill in the name and text fields. */
	bp -> mst_name  = str_alloc(symbol, OBJ_MACRO_STRING);
	bp -> mst_text  = str_alloc(txt, OBJ_MACRO_RTEXT);
	bp -> mst_nargs = nargs;
	
	/* Add it to the proper hash list. */
	lst_add(bp, &mst_ht[hash]);
}

/*
	Return the hash value for symbol.
*/
static int
mst_hash (reg char * symbol)
{
	reg int hash;

	STAT("mst_hash");

	for (hash = 0; *symbol; ) {
		hash *= 3;
		hash += (int) *symbol++;
		hash %= MAC_PRIME;
	}
	return hash;
}

/*
	Initialize the macro module.
	This must be done BEFORE command line arguments are processed.
*/
void
mst_init(void)
{
	reg int i;

	TICK("mst_init");

	/* Clear the hash table. */
	for (i = 0; i < MAC_PRIME; i++) {
		mst_ht [i] = NULL;
	}

	/* Clear the "initial undef list." */
	undef_list = NULL;
}

/*
	Enter the predefined macros into the table.
	Do nothing if the initial definitions have been
	suppressed with -u options.

	This must be done AFTER command line options have been processed.
*/
void
mst2_init(void)
{
	mst_enter("__LINE__", "", -3);
	mst_enter("__FILE__", "", -3);
	mst_enter("__STDC__", "", -3);
	mst_enter("__TIME__", "", -3);
	mst_enter("__DATE__", "", -3);
}

/*
	Delete all macro entries, then enter predefined macros back into the table.
*/
void
mst_init_again(void)
{
	reg int			i;
	reg mst_node *	np;

	TICKB("mst_init_again");
	
	/* Free all macro definitions. */
	for (i = 0; i < MAC_PRIME; i++) {
		lst_every(mst_node, np, &mst_ht[i]) {
			obj_free(np -> mst_name);
			obj_free(np -> mst_text);
		}
		lst_free_all(&mst_ht[i]);
	}
	
	/* Erase the undef list. */
	lst_free_all(&undef_list);
	
	/*
		Do *not* restore the predefined macros.
		They will be restored by mst_restore unless
		they were over-ridden on the command line.
	*/
#if 0
	mst2_init();
#endif
	
	TICKX("mst_init_again");
}

/*
	Look up a symbol in the macro table.
	Return TRUE if found, and set text and nargs.
*/
bool
mst_lookup(reg char * symbol, char * VAR text, int VAR nargs)
{
	FTAG("mst_lookup");

	reg mst_node *	bp;
	reg int hash = mst_hash(symbol);

	TRACEP(ftag, es(symbol));
	
	lst_every(mst_node, bp, &mst_ht[hash]) {
		if (str_eq(symbol, bp -> mst_name)) {

			/* Special case for __line__ and __file__. */
			if (bp -> mst_nargs == -3) {
				if (str_eq(bp -> mst_name, "__LINE__")) {
				
					/* Set current line number. */
					char buffer [30];
					bp -> mst_text =
						str_alloc(cvt_long(buffer, 30, err_line), OBJ_MACRO_LINE);
				}
				else if (str_eq(bp -> mst_name, "__FILE__")) {

					/* Set current file name. */
					char buffer [200];
					bp -> mst_text =
						str_alloc(cvt_quote(buffer, 200, err_file_name()), OBJ_MACRO_FILE);
				}
				else if (str_eq(bp -> mst_name, "__TIME__")) {
					/* Set current file name. */
					bp -> mst_text =
						str_alloc(the_time(), OBJ_MACRO_TIME);
				}
				else if (str_eq(bp -> mst_name, "__DATE__")) {
					bp -> mst_text =
						str_alloc(the_date(), OBJ_MACRO_DATE);
				}
				else if (str_eq(bp -> mst_name, "__STDC__")) {
					bp -> mst_text = "1";
				}
			}

			/* Return success. */
			*nargs = max(-1, bp -> mst_nargs);
			*text  = bp -> mst_text;

			TRACEN(ftag,
				es(" found: <"); es_str(*text); es(">\n"); enl());
				
			return TRUE;
		}
	}

	TRACEN("mst_lookup", es(" not found\n"));

	/* Return failure. */
	return FALSE;
}

/*
	Restore the  macro table to the defines and undefs that were
	entered from the previous command-line.
*/
void
mst_restore (void)
{
	reg unode *		up;
	reg mst_node *	np;
	reg int i;
	
	TRACEP("mst_restore",
		lst_report(&mst_udump, &undef_save, "undef save list");
		lst_report(&mst_ndump, &def_save, "macro save list"));
	
	/* Restore the undef list. */
	lst_free_all(&undef_list);
	lst_every(unode, up, &undef_save) {
		mst_undef_arg(up -> u_name);
	}
	lst_free_all(&undef_save);
	
	/* Restore the symbol table itself. */
	for (i = 0; i < MAC_PRIME; i++) {
		mst_ht[i] = NULL;
	}
	lst_every(mst_node, np, &def_save) {
		mst_enter(np -> mst_name, np -> mst_text, np -> mst_nargs);
	}
	lst_free_all(&def_save);
}

/*
	Save the current macro table so that it can be restored by mst_restore().
*/
void
mst_save (void)
{
	reg unode *		up;
	reg unode *		up2;
	reg mst_node *	np;
	reg mst_node *	np2;
	reg int i;
	
	SL_DISABLE();
	
	/* Create a copy of the undef list. */
	lst_free_all(&undef_save);
	lst_every(unode, up, &undef_list) {
		up2 = obj_new(unode, OBJ_UNDEF_STRUCT);
		up2 -> u_name = up -> u_name;
		up2 -> u_flag = TRUE;
		lst_append(up2, &undef_save);
	}
	
	/* Create a copy of the symbol table. */
	lst_free_all(&def_save);
	for (i = 0; i < MAC_PRIME; i++) {
		lst_every(mst_node, np, &mst_ht[i]) {
		
			/* Bug fix: 4/6/96: we _do_ have to call str_alloc here! */
			np2 = obj_new(mst_node, OBJ_MST_STRUCT+M_NODUMP);
			np2 -> mst_name  = str_alloc(np -> mst_name, OBJ_MACRO_STRING);
			np2 -> mst_text  = str_alloc(np -> mst_text, OBJ_MACRO_RTEXT);
			np2 -> mst_nargs = np -> mst_nargs;
			lst_append(np2, &def_save);
		}
	}
	
	TRACEP("mst_save",
		lst_report(&mst_udump, &undef_save, "undef save list");
		lst_report(&mst_ndump, &def_save, "macro save list"));
}

/*
	Make a simple test of this module.
	This is intended only to catch obvious blunders.
*/
void
mst_test(void)
{
	char * rtext;
	int  nargs;
	
	SL_DISABLE();

	mst_lookup("__DATE__", &rtext, &nargs);
	es("mst_test: date: "); es(rtext); ecs();
	
	mst_lookup("__TIME__", &rtext, &nargs);
	es("time: "); es(rtext); ecs();
	
	mst_lookup("__FILE__", &rtext, &nargs);
	es("file: "); es(rtext); ecs();
	
	mst_lookup("__STDC__", &rtext, &nargs);
	es("STDC: "); es(rtext); enl();
}

/*
	Add an argument to the "initial undef list."
*/
void
mst_undef_arg(char * arg)
{
	unode * p;
	
	SL_DISABLE();
	ASSERT(arg);
	
	TRACEP("mst_undef_arg", es(arg); enl());

	p = obj_new(unode, OBJ_UNDEF_STRUCT);
	p -> u_name	= arg;
	p -> u_flag = TRUE;
	lst_add(p, &undef_list);
}

/*
	Print one mst_node.
*/
static void
mst_ndump (lst_node * lnp)
{
	mst_node * np = (mst_node *) lnp;
	
	es("name: ");		es(np -> mst_name);
	es(" nargs: ");		eint(np -> mst_nargs);
	es(" rtext: <");	es_str(np -> mst_text);
	es("> at: ");		eptr(np); enl();
}

/*
	Print one unode.
*/
static void
mst_udump (lst_node * lnp)
{
	unode * up = (unode *) lnp;
	
	es("name: ");	es(up -> u_name);
	es(" flag: ");	ebool(up -> u_flag);
	es(" at: ");	eptr(up); enl();
}
