/*
	SPP: Statement Parser.

	source:  SPPpar.c
	started: October 22, 1985
	version:
		March 22, 1996.
			Added copy_buf logic.
		March 17, 1996.
*/

#include "SPP.h"

#include "SPPdcl.h"
#include "SPPexp.h"
#include "SPPhold.h"
#include "SPPpar.h"
#include "SPPpr.h"
#include "SPPmain.h"
#include "SPPsem.h"
#include "SPPst.h"
#include "SPPtok.h"

#include "SPPenv.h"	/* Library additions. */
#include "SPPerr.h"
#include "SPPlst.h"
#include "SPPobj.h"
#include "SPPstr.h"

/*
	Function prototypes of internal parsing routines.
*/
static void	stat_list	(void);
static bool	stat1		(void);
static void do_asm		(void);
static void	do_if		(void);
static void	do_do		(void);
static void	do_while	(void);
static void	do_do		(void);
static void	do_while	(void);
static void	do_for		(void);
static void	do_switch	(void);
static void	do_break	(void);
static void	do_continue	(void);
static void	do_case		(void);
static void	do_default	(void);
static void	do_return	(void);
static void	do_goto		(void);
static void	do_label	(void);
static void	show_need	(char * s);

/*
	Parse a block, including opening and closing braces.
	This is NOT called to parse the block that comprises a function body.
*/
void
block(void)
{
	TRACETOK("block");

	must(LCURLY_TOK);
	block_dcls();
	stat_list();
	need(RCURLY_TOK);

	TICKX("block");
}

/*
	Parse a function body, including opening and closing braces.
*/
void
body(void)
{
	TRACETOK("body");

	must(LCURLY_TOK);
	
	/* Output the FTAG or SL_NAME macros. */
	so_ftag();

	/* All future definitions are local to the block. */
	st_begin();

	/* Parse local declarations. */
	block_dcls();

	/* Output the sherlock macro. */
	so_entry();
	
	sf_1body();
	stat_list();
	sf_2body();

	/* Output a TICKX() macro if we have fallen through the end. */
	so_leave();

	need(RCURLY_TOK);

	/* Go back to global definitions. */
	st_end();

	TICKX("body");
}

/*
	Skip the current token if it is what we think it is.
*/
void
need(en_tokens tok)
{
	TRACEPB("need", es_op(tok); enl());

	if (token == tok) {
		get_token();
	}
	else {
		show_need(pr_op(tok));
	}

	TICKX("need");
}

/*
	Internal error if the token is not found.
	Othewise, eat it.
*/
void
must(en_tokens expected)
{
	TRACEPB("must", es_op(expected); enl());

	if (token != expected) {
	
		fatal(
			es("must: <"); es_op(expected);
			es("> expected, <"); es_op(token); es("> found\n"));
	}
	else {
		get_token();
	}

	TICKX("must");
}

/*
	error message about needed symbol
*/
static void
show_need(char * s)
{
	TRACEPB("show_need", es(s); enl());
	
	error(
		es(s); es(" expected");
		if (token == ID_TOK) {
			es(" at: "); es(t_symbol);
		}
		else if (pr_op(token)) {
			es(" at: "); es_op(token);
		}
		enl());

	TICKX("show_need");
}

/*
	If the current token is tok, return TRUE and eat it.

	Otherwise find the next instance of tok, eat it, and return FALSE.
	-- BUT --
	Stop on and return an intervening semicolon or brace.
*/

bool
needend(en_tokens tok)
{
	TRACETOK("needend");

	if (token == tok) {
		get_token();

		STATX("needend");
		return TRUE;
	}
	show_need(pr_op(tok));

	for(;;) {
		switch(token) {
		case SEMICOLON_TOK:
		case LCURLY_TOK:
		case RCURLY_TOK:
		case EOP_TOK:
			STATX("needend");
			return FALSE;
		}
		get_token();
		if (token == tok) {
			get_token();
			STATX("needend");
			return FALSE;
		}
	}
}

/*
	Parse a statement list.
*/
static void
stat_list(void)
{
	TRACETOK("stat_list");

	while(stat1()) {
		;
	}

	TICKX("stat_list");
}
	
/*
	Parse a single statement.
	Return TRUE if a statement was seen.
*/
static bool
stat1(void)
{
	TRACETOK("stat1");

loop:

	switch(token) {

	case EOP_TOK:
	case RCURLY_TOK:
		STATX("stat1");
		return FALSE;

	case LCURLY_TOK:
		block();
		STATX("stat1");
		return TRUE;

	case SEMICOLON_TOK:
		get_token();
		goto loop;

	case K_BREAK:	 do_break();		STATX("stat1"); return TRUE;
	case K_CASE:	 do_case();			STATX("stat1"); return TRUE;
	case K_CONTINUE: do_continue();		STATX("stat1"); return TRUE;
	case K_DEFAULT:	 do_default();		STATX("stat1"); return TRUE;
	case K_DO:	 	 do_do();			STATX("stat1"); return TRUE;
	case K_FOR:	 	 do_for();			STATX("stat1"); return TRUE;
	case K_GOTO:	 do_goto();			STATX("stat1"); return TRUE;
	case K_IF:	 	 do_if();			STATX("stat1"); return TRUE;
	case K_SWITCH:	 do_switch();		STATX("stat1"); return TRUE;
	case K_WHILE:	 do_while();		STATX("stat1"); return TRUE;
	case K_RETURN:	 do_return();		STATX("stat1"); return TRUE;

	case ID_TOK:
	
#ifdef ENV_LSC
		if (str_eq(t_symbol, "asm")) {
			do_asm();
			STATX("stat1");
			return TRUE;
		}
#endif
		
		/* Do character-oriented lookahead. */
		skip_bl();
		if (ch == ':') {
			do_label();
			goto loop;
		}

	default:
		if (is_type()) {
			error(es("Declaration in code: check brace structure"));
			STATX("stat1");
			return FALSE;
		}
		else if (is_expr_tok()) {
			sf_expr();
			expr(SEMICOLON_TOK);
			need(SEMICOLON_TOK);
			STATX("stat1");
			return TRUE;
		}
		else {
			STATX("stat1");
			return FALSE;
		}
	}
}

/*
	Parse the lightspeed asm { ... } statement.
*/
static void
do_asm(void)
{
	TRACETOK("do_asm");
	
	must(ID_TOK);
	must(LCURLY_TOK);
	while (token != RCURLY_TOK) {
		get_token();
	}
	must(RCURLY_TOK);
	
	TICKX("do_asm");
}

/*
	Parse complex statements.
*/
static void
do_do(void)
{
	int oldcont;
	int oldbrk;

	TRACETOK("do_do");

	oldcont  = cont_ok;		cont_ok  = TRUE;
	oldbrk   = break_ok;	break_ok = TRUE;

	must(K_DO);
	sf_1do();
	stat1();
	sf_2do();
	need(K_WHILE);
	need(LPAREN_TOK);
	expr(RPAREN_TOK);
	need(RPAREN_TOK);

	cont_ok  = oldcont;
	break_ok = oldbrk;

	TICKX("do_do");
}

static void
do_for(void)
{
	int oldcont;
	int oldbrk;

	TRACETOK("do_for");

	oldcont = cont_ok;	cont_ok  = TRUE;
	oldbrk  = break_ok;	break_ok = TRUE;

	must(K_FOR);
	need(LPAREN_TOK);
	expr(SEMICOLON_TOK);	need(SEMICOLON_TOK);
	sf_1for(is(SEMICOLON_TOK));
	expr(SEMICOLON_TOK);	need(SEMICOLON_TOK);
	expr(RPAREN_TOK);		need(RPAREN_TOK);
	stat1();
	sf_2for();

	cont_ok  = oldcont;
	break_ok = oldbrk;

	TICKX("do_for");
}

static void
do_if(void)
{
	TRACETOK("do_if");

	must(K_IF);
	need(LPAREN_TOK);
	expr(RPAREN_TOK);
	need(RPAREN_TOK);
	sf_1if();
	stat1();
	if (token == K_ELSE) {
		must(K_ELSE);
		sf_2if();
		stat1();
	}
	sf_3if();

	TICKX("do_if");
}

static void
do_switch(void)
{
	int oldswitch;
	int oldbrk;

	TRACETOK("do_switch");
             
	oldswitch = in_switch;	in_switch = TRUE;
	oldbrk    = break_ok;	break_ok  = TRUE;

	must(K_SWITCH);
	need(LPAREN_TOK);
	expr(RPAREN_TOK);
	need(RPAREN_TOK);
	need(LCURLY_TOK);
	sf_1switch();
	stat_list();
	sf_2switch();
	need(RCURLY_TOK);

	in_switch = oldswitch;
	break_ok  = oldbrk;

	TICKX("do_switch");
}

static void
do_while(void)
{
	int oldcont;
	int oldbrk;

	TRACETOK("do_while");

	oldcont = cont_ok;	cont_ok  = TRUE;
	oldbrk  = break_ok;	break_ok = TRUE;

	must(K_WHILE);
	need(LPAREN_TOK);
	expr(RPAREN_TOK);
	need(RPAREN_TOK);
	sf_1while();
	stat1();
	sf_2while();

	cont_ok  = oldcont;
	break_ok = oldbrk;

	TICKX("do_while");
}

/*
	Parse subordinate statements.
*/
static void
do_break(void)
{
	TRACETOK("do_break");

	if (!break_ok) {
		error(es("'break' does not match any statement"));
	}

	sf_break();
	must(K_BREAK);
	need(SEMICOLON_TOK);

	TICKX("do_break");
}

static void
do_case(void)
{
	TRACETOK("do_case");

	if (!in_switch) {
		error(es("'case' appears outside of any switch"));
	}

	must(K_CASE);
	con_expr(COLON_TOK);
	need(COLON_TOK);

	TICKX("do_case");
}
		
static void
do_continue(void)
{
	TRACETOK("do_continue");

	if (!cont_ok) {
		error(es("'continue' appears outside do, while and if"));
	}

	must(K_CONTINUE);
	need(SEMICOLON_TOK);

	TICKX("do_continue");
}

static void
do_default(void)
{
	TRACETOK("do_default");

	/* Make sure we are in a switch statement. */
	if (!in_switch) {
		error(es("'default' appears outside of a switch"));
	}

	sf_default();
	must(K_DEFAULT);
	need(COLON_TOK);

	TICKX("do_default");
}

static void
do_goto(void)
{
	TRACETOK("do_goto");

	sf_goto();
	must(K_GOTO);
	need(ID_TOK);
	needend(SEMICOLON_TOK);

	TICKX("do_goto");
}

static void
do_label(void)
{
	TICKB("do_label");


	sf_label();
	must(ID_TOK);
	need(COLON_TOK);

	TICKX("do_label");
}

static void
do_return(void)
{
	TRACETOK("do_return");

	sf_return();
	so_1exit();
	
	/* Skip the return, and start buffering now. */
	ASSERT(token == K_RETURN);
	copy_buf_flag = TRUE;
	copy_count = 0;
	get_token();
	
	if (token == SEMICOLON_TOK) {
		copy_buf_flag = FALSE;
		so_2exit(FALSE);
	}
	else {
		/* Put the return expression into copy_buf []. */
		so_2exit(TRUE);
		expr(SEMICOLON_TOK);
	}
	so_3exit();
	copy_count = 0;
	copy_buf_flag = FALSE;
	need(SEMICOLON_TOK);

	TICKX("do_return");
}
