/*
	SPP: semantic routines.
	
	Source: SPPsem.c
	Started: October 1, 1987
	Version:
		March 22, 1996.
			Created so_tracepx.
		March 21, 1996.
			Created so_return().
		March 17, 1996.
			Created so_ftag(), so_exxx() and so_fname().
*/

#include "SPP.h"

#include "SPParg.h"
#include "SPPdcl.h"
#include "SPPdef.h"
#include "SPPhold.h"
#include "SPPpr.h"
#include "SPPres.h"
#include "SPPsem.h"
#include "SPPst.h"
#include "SPPtok.h"

#include "SPPerr.h"	/* Library additions. */
#include "SPPio.h"
#include "SPPobj.h"
#include "SPPstr.h"

#include <LIBcvt.h>

#define TRACE_SCOPE(name) TRACEP(name, es_scope(scope); ecnl());

/*
	These abbreviations help keep the paren balancer happy.
*/
#define open_paren "("
#define close_paren ")"
#define tab "\t"
#define double_quote "\""

/*
	Information about the declaration currently being parsed.

	cur_fid is the "default" holding place for identifiers.
	cur_pid is used for identifiers in parameter lists.

	cur_hxtype is the type of the common head for funcs or params.
*/
static bool	 cur_typedef = FALSE;		/* TRUE if typedef seen.	*/

static en_tokens cur_hptype = K_INT;	/* Common head param type.	*/
static en_tokens cur_hftype = K_INT;	/* Common head func type.	*/

static en_tokens cur_ptype = NULL_TOK;	/* Effective param type.	*/
static en_tokens cur_ftype = NULL_TOK;	/* Effective func  type.	*/

#define MAX_ID 100

static char		cur_pid [MAX_ID];		/* Defined param id.		*/
static char		cur_fid	[MAX_ID];		/* Current (func) id.		*/
static char		fn_name	[MAX_ID];		/* Current function name.	*/

/*
	Information about the function whose body is being parsed.
*/
static en_tokens	fn_type = NULL_TOK;	/* Effective printf type.		*/
static bool	 		fn_enable;			/* Enable generation of macros.	*/
static bool	 		fn_gen;				/* Macros generated flag.		*/
static bool	 		fn_warn;			/* Warning generated flag.		*/

static struct param {
	char * 		p_name;
	en_tokens	p_type;
};

/*
	Define the table of formal parameters.
	This table is created in NEW_PARAM_SCOPE and OLD_PARAM_SCOPE.
	Entries in the table are make by sd_comma() and sd_end().
*/

#define MAX_FPARAM 100

static struct param pst [MAX_FPARAM];
static int n_param = 0;

/*
	Forward declarations of functions.
*/
static en_tokens	new_type	(en_tokens old_tok, en_tokens new_tok);

static void sd_dump		(void);
static void so_exxx		(void);
static void so_fname	(void);
static void so_main		(void);
static void so_printf	(int start, int end);
static void	so_return	(void);
static void	so_tracepx1	(void);
static void	so_tracepx2	(void);
static void so_usr		(void);

static char *		usr2_cvt	(en_tokens type);

/*
	Return TRUE if tok is a base type or pointer or function.

	2/10/89 bug fix:

	If this routine does not allow STRING_TOK as a base type, then
	the special casing of typedef char * charp does not work properly.
*/
bool
is_base(en_tokens tok)
{
	SL_DISABLE();
	
	return	tok == K_CHAR   || tok == K_DOUBLE ||
		tok == K_FLOAT  	|| tok == K_INT    ||
		tok == K_LONG   	|| tok == K_SHORT  ||
		tok == K_STRUCT 	|| tok == K_UNION  ||
		tok == K_VOID  		|| tok == K_ENUM   ||
		tok == K_UNSIGNED 	|| tok == K_SIGNED ||
		tok == STAR_TOK 	||
		tok == STRING_TOK 	||		/* bug fix: 2/10/89 */
		tok == UNSIGNED_LONG ||		/* bug fix: 2/10/89 */
		tok == K_BOOL 		||		/* bug fix: 2/17/89 */
		tok == LPAREN_TOK	|| tok == LBRACK_TOK;
}

/*
	Return TRUE if the current token is either a base type or
	a typedef'd id.
*/
bool
is_type(void)
{
	FTAG("is_type");
	TRACEPB(ftag, es("tok: "); es_tok(); ecnl());

	/* Allow local keywords here. */
	allow_mkeys();
	if (token == ID_TOK) {
		STATX(ftag);
		return is_typedef(t_symbol);
	}
	else {
		STATX(ftag);
		return is_kdecl(token);
	}
}

/*
	Return TRUE if id is a typedef'd id.
*/
bool
is_typedef(char * id)
{
	FTAG("is_typedef");

	en_tokens type;
	bool td_flag;

	TRACEP(ftag, es(id); enl());

	if (st_lookup(id, &type, &td_flag) == FALSE) {
		TRACEPN(ftag, eret(); es("FALSE\n"));
		return FALSE;
	}
	else {
		TRACEPN(ftag, eret(); eint(td_flag); enl());
		return td_flag;
	}
}

/*
	Compute the new print type, based on the old and new tokens.

	OLD			NEW		RESULT
	unsigned	long	unsigned_long
	unsigned	int		unsigned
	unsigned	short	unsigned
	unsigned	char	char
	long		int		long
	char		*		string
	any			(		old_tok
*/
static en_tokens
new_type(en_tokens old_tok, en_tokens new_tok)
{
	FTAG("new_type");

	en_tokens tok;

	TRACEPB(ftag, es_op(old_tok); ecs(); es_op(new_tok); enl());

	if (new_tok == LPAREN_TOK) {
		tok = old_tok;
	}
	else if (old_tok == K_CHAR && new_tok == STAR_TOK) {
		tok = STRING_TOK;
	}
	/* Bug fix: 9/14/90 */
	else if (new_tok == STAR_TOK) {
		tok = STAR_TOK;
	}
	else if (old_tok == UNSIGNED_LONG) {
		/* unsigned long and unsigned long int. */
		tok = UNSIGNED_LONG;
	}
	else if (old_tok == K_UNSIGNED && new_tok == K_LONG) {
		tok = UNSIGNED_LONG;
	}
	else if (old_tok == K_UNSIGNED && new_tok == K_CHAR) {
		tok = K_CHAR;
	}
	else if (old_tok == K_UNSIGNED) {
		/* unsigned int or unsigned short. */
		tok = K_UNSIGNED;
	}
	else if (old_tok == K_LONG && new_tok == K_INT) {
		tok = K_LONG;
	}
	else {	
		tok = new_tok;
	}
	TRACEPX(ftag, eret(); es_op(tok); enl());
	return tok;
}

/*
	D e c l a r a t i o n    R o u t i n e s.
*/

/*
	Begin/end a declaration.
*/
void
sd_dcl(en_scope scope)
{
	TRACE_SCOPE("sd_dcl");

	if (scope == OUTER_SCOPE) {
		int i;
		for(i = 0; i < n_param; i++) {
			obj_free(pst [i] . p_name);
		}
		n_param    = 0;
		fn_type    = NULL_TOK;
		fn_name [0] = '\0';
	}
	if (scope == OUTER_SCOPE || scope == BLOCK_SCOPE) {
		cur_typedef = FALSE;
		cur_hptype = K_INT;
		cur_hftype = K_INT;
		cur_ptype  = NULL_TOK;
		cur_ftype  = NULL_TOK;
		cur_pid [0] = '\0';
		cur_fid [0] = '\0';
	}
}

void
sd_end(en_scope scope, en_tokens end_tok)
{
	FTAG("sd_end");

	int i;

	TRACEP("v", es_scope(scope); ecs(); es_op(end_tok); enl());

	if (scope == OUTER_SCOPE || scope == BLOCK_SCOPE) {
		if (cur_typedef) {

			TRACEPN(ftag, es("*****typedef id:  "); es(cur_fid); enl());

			/* 2/27/89: Convert typedef int bool to K_BOOL */
			if (cur_ftype == K_INT && str_eq(cur_fid, "bool")) {
				cur_ftype = K_BOOL;
			}
			st_enter(cur_fid, cur_ftype, TRUE);
		}
		if (end_tok == COMMA_TOK) {
			cur_fid [0] = '\0';
			cur_ftype = cur_hftype;
		}

		/* bug fix: 2/14/89 */
		cur_hftype = K_INT;
	}
	else if (scope == NEW_FORMAL_SCOPE) {
		if (!str_eq(cur_pid, "")) {

			TRACEPN(ftag,
				es("new formal: "); es_op(cur_ptype); ecs(); es(cur_pid); enl());

			/* Enter the id into list of formal parameters. */
			pst [n_param] . p_name = str_alloc(cur_pid, OBJ_PARAM_STRING);
			pst [n_param] . p_type = cur_ptype;
			n_param++;
		}
		
		/* There is no common type in NEW_FORMAL_SCOPE. */
		cur_pid [0] = '\0';
		cur_ptype = NULL_TOK;

		/* bug fix: 2/14/89 */
		cur_hptype = K_INT;
	}
	else if (scope == OLD_FORMAL_SCOPE) {
		if (!str_eq(cur_pid, "")) {
			/* Find the variable on the list of formals. */
			for (i = 0; i < n_param; i++) {
				if (str_eq(cur_pid, pst[i].p_name)) {
					break;
				}
			}
			if (i == n_param) {
				error(es("Missing from formal param list: "); es(cur_pid); enl());
			}
			else if (pst[i]. p_type != NULL_TOK) {
				error(es("Duplicate declaration of: "); es(cur_pid); enl());
			}
			else {
				TRACEPN(ftag,
					es("old formal: "); es_op(cur_ptype); ecs(); es(cur_pid); enl());
				
				pst[i].p_type = cur_ptype;
			}
		}
		if (end_tok == COMMA_TOK) {
			cur_pid [0] = '\0';
			cur_ptype = cur_hptype;
		}

		/* bug fix: 2/14/89 */
		else {
			cur_hptype = K_INT;
		}
	}
}

/*
	Set the name of the current function.
*/
void
sd_fn_name(char * name)
{
	SL_DISABLE();

	str_ncpy(fn_name, name, MAX_ID);
}

/*
	Begin a function definition.
	This is called just after the opening curly bracket of the function.
*/
void
sd_func(void)
{
	FTAG("sd_func");

	int i;
	
	SL_DISABLE();

	/* Save the current settings. */
	str_ncpy(fn_name, cur_fid, MAX_ID);
	
	/* Set name of function for error messages. */
	err_begin_func(fn_name);
	
	if (cur_ftype == NULL_TOK) {
		fn_type = cur_hftype;
	}
	else {
		fn_type = cur_ftype;
	}

	/* 2/16/89:
		For this function:
		1. enable generation of Sherlock macros and
		2. indicate that no Sherlock macro has been generated yet.
		3. indicate that no warning has been generated yet. 
	*/
	fn_enable = TRUE;
	fn_gen    = FALSE;
	fn_warn   = FALSE;

	/*	2/14/89
		Give error message for undefined formal types.
	*/
	for (i = 0; i < n_param; i++) {
		if (pst[i].p_type == NULL_TOK) {
			warning(
				es("Missing declaration for formal parameter: ");
				es(pst[i].p_name); enl());
		}
	}

	TRACEP(ftag,
		es("function: "); es(fn_name); es(" type: "); es_op(fn_type); ecnl();
		for (i = 0; i < n_param; i++) {
			es("param "); eint(i);
			es(": "); es(pst[i].p_name);
			es(" type: "); es_op(pst[i].p_type); enl();
		};
	);
}

/* Set the common header type of the current (sub) declaration. */
void
sd_htype(en_scope scope, en_tokens tok)
{
	FTAG("sd_htype");

	bool flag;
	
	SL_DISABLE();

	if (tok == ID_TOK) {
		/* Get the type from an already defined typedef id. */
		st_lookup(t_symbol, &tok, &flag);
	}
	if (!is_base(tok)) {
		return;
	}

	TRACEP(ftag, es_scope(scope); ecs(); es_op(tok); enl());

	if (scope == OUTER_SCOPE || scope == BLOCK_SCOPE) {
		/* Set function type. */
		if (tok != LPAREN_TOK) {
			/* bug fixes:  2/14/89 and 12/11/90 */
			if (!flag) {
				tok = new_type(cur_hftype, tok);
			}
			cur_hftype = tok;
			cur_ftype  = tok;
		}
	}
	else if (scope == NEW_FORMAL_SCOPE || scope == OLD_FORMAL_SCOPE) {

		/* Set formal type. */
		/* bug fix: 12/11/90 */
		if (!flag) {
			tok = new_type(cur_hptype, tok);
		}
		cur_hptype = tok;
		cur_ptype  = tok;
	}
}

/*
	Set the type of the current (sub) declaration.
	Token may be LPAREN_TOK or LBRACK_TOK.
*/
void
sd_type(en_scope scope, en_tokens tok)
{
	FTAG("sd_type");

	bool 		garbage;
	en_tokens	old_tok;
	
	SL_DISABLE();

	if (tok == ID_TOK) {
		/* Get the type from an already defined typedef id. */
		st_lookup(t_symbol, &tok, &garbage);
	}
	if (!is_base(tok)) {
		return;
	}

	TRACEP(ftag, es_scope(scope); ecs(); es_op(tok); enl());

	if (scope == OUTER_SCOPE || scope == BLOCK_SCOPE) {

		old_tok = (cur_ftype == NULL_TOK) ? cur_hftype : cur_ftype;

		/* bug fix:  2/14/89 */
		cur_ftype = new_type(old_tok, tok);
	}
	else if (scope == NEW_FORMAL_SCOPE || scope == OLD_FORMAL_SCOPE) {

		old_tok = (cur_ptype == NULL_TOK) ? cur_hptype : cur_ptype;

		/* bug fix: 2/14/89 */
		cur_ptype = new_type(old_tok, tok);
	}
}

/*
	Set the typedef flag.
*/
void
sd_typedef(en_scope scope)
{
	TRACE_SCOPE("sd_typedef");

	if(scope == OUTER_SCOPE || scope == BLOCK_SCOPE) {
		cur_typedef = TRUE;
	}
}

/*
	Handle an identifier.
*/
void
sd_id(en_scope scope, char * id)
{
	TRACEP("sd_id", es_scope(scope); es(id); enl());

	if (scope == OUTER_SCOPE || scope == BLOCK_SCOPE) {
		str_ncpy(cur_fid, id, MAX_ID);
	}
	else if (scope == NEW_FORMAL_SCOPE || scope == OLD_FORMAL_SCOPE) {
		str_ncpy(cur_pid, id, MAX_ID);
	}
}

/*
	R e - i n i t i a l i z a t i o n    R o u t i n e s.
*/

/*
	Reinitialize this module.
*/
void
sem_init_again (void)
{
	int i;
	
	cur_typedef	= FALSE;		/* TRUE if typedef seen.	*/
	cur_hptype	= K_INT;		/* Common head param type.	*/
	cur_hftype	= K_INT;		/* Common head func type.	*/
	cur_ptype	= NULL_TOK;		/* Effective param type.	*/
	cur_ftype	= NULL_TOK;		/* Effective func  type.	*/

	cur_pid	[0] = '\0';			/* Defined param id.		*/
	cur_fid	[0] = '\0';			/* Current (func) id.		*/
	fn_name	[0] = '\0';			/* Current function name.	*/
	
	fn_type		= NULL_TOK;		/* Effective printf type.	*/
	fn_enable	= TRUE;			/* Enable generation of macros.	*/
	fn_gen		= FALSE;		/* Macros generated flag.		*/
	fn_warn		= FALSE;		/* Warning generated flag.		*/
	
	for(i = 0; i < n_param; i++) {
		obj_free(pst [i] . p_name);
	}
	n_param	= 0;
}

/*
	F l o w    C o n t r o l    R o u t i n e s.
*/

/*
	Define format of entries in flow control stack.
*/
typedef struct flow_struct flow_node;

static struct flow_struct {
	en_tokens	f_type;		/* Statement type.		*/
	int			f_status;	/* Termination info.	*/
};

/* Define bits in f_term field of flow_node. */

#define FALL_BIT		0x01	/* 1 if WILL fall through.	*/
#define ELSE_BIT		0x02	/* 1 if else clause seen.	*/
#define DEFAULT_BIT		0x04	/* 1 if default seen.		*/
#define BREAK_BIT		0x08	/* 1 if break seen.			*/
#define NULL_FOR_BIT	0x10	/* 1 if empty for control.	*/

#define MAX_FLOW 100
static flow_node	sf_nodes [MAX_FLOW];
static int			sf_level;		/* Current nesting level.	*/

/*
	Initialize the flow control.
	An empty body will get a TICKX macro.
*/
void
sf_1body(void)
{
	flow_node * p;

	SL_DISABLE();
	
	sf_level = 0;
	p = & sf_nodes [sf_level];
	p -> f_type   = NULL_TOK;
	p -> f_status = FALL_BIT;
}

void
sf_2body(void)
{
	SL_DISABLE();
	
	if (sf_level != 0) {
		fatal(es("sf_2body: can't happen"));
	}
}

void
sf_break(void)
{
	int 		i;
	en_tokens	type;
	
	SL_DISABLE();

	/* Set flag showing that the enclosing loop does NOT terminate. */
	for (i = sf_level; i >= 0; i--) {
		type = sf_nodes [i] . f_type;
		if (type == K_DO || type == K_FOR ||
		    type == K_WHILE || type == K_SWITCH) {
			sf_nodes [i].f_status |= BREAK_BIT;
			break;
		}
		else if (type == K_IF) {
			sf_nodes [i].f_status |= BREAK_BIT;
		}
	}
}

void
sf_1do(void)
{
	flow_node * p;
	
	SL_DISABLE();

	if (sf_level >= MAX_FLOW) {
		warning(es("Statements nested too deeply"); enl());
	}
	else {
		p = & sf_nodes [++sf_level];
		p -> f_type   = K_DO;
		p -> f_status = 0;
	}
}

void
sf_2do(void)
{
	SL_DISABLE();
	
	sf_level--;
	sf_nodes [sf_level].f_status |= FALL_BIT;	/* Never terminates. */
}

void
sf_expr(void)
{
	SL_DISABLE();
	
	sf_nodes [sf_level].f_status |= FALL_BIT;	/* Never terminates. */
}

void
sf_1for(bool empty_flag)
{
	flow_node * p;
	
	SL_DISABLE();

	if (sf_level >= MAX_FLOW) {
		warning(es("Statements nested too deeply"); enl());
	}
	else {
		p = & sf_nodes [++sf_level];
		p -> f_type   = K_DO;
		p -> f_status = 0;
		if (empty_flag) {
			p -> f_status |= NULL_FOR_BIT;
		}
	}
}

void
sf_2for(void)
{
	flow_node *p;
	
	SL_DISABLE();

	p = & sf_nodes [sf_level];
	sf_level--;
	if (	(p -> f_status & BREAK_BIT) ||
		(p -> f_status & NULL_FOR_BIT) == 0) {
		sf_nodes [sf_level].f_status |= FALL_BIT;
	}
	else {
		sf_nodes [sf_level].f_status &= ~FALL_BIT;
	}
}

void
sf_goto(void)
{
	SL_DISABLE();
	
	/* goto's NEVER fall through, but labels ALWAYS do. */
	sf_nodes [sf_level].f_status &= ~FALL_BIT;
}

void
sf_1if(void)
{
	flow_node * p;
	
	SL_DISABLE();

	if (sf_level >= MAX_FLOW) {
		warning(es("Statements nested too deeply"); enl());
	}
	else {
		p = & sf_nodes [++sf_level];
		p -> f_type   = K_IF;
		p -> f_status = 0;
	}
}

void
sf_2if(void)
{
	flow_node * p;
	
	SL_DISABLE();

	/*
		Only set ELSE_BIT if the 'then' clause terminated.
	*/
	p = & sf_nodes [sf_level];
	if ( (p -> f_status & FALL_BIT) == 0) {
		p -> f_status |= ELSE_BIT;
	}
}

void
sf_3if(void)
{
	flow_node * p;
	
	SL_DISABLE();

	/* Both braches must have terminated. */
	p = & sf_nodes [sf_level];
	sf_level--;
	if (	(p -> f_status & ELSE_BIT) == 0 ||
		(p -> f_status & BREAK_BIT) ||
		(p -> f_status & FALL_BIT)) {
		sf_nodes [sf_level].f_status |= FALL_BIT;
	}
	else {
		sf_nodes [sf_level].f_status &= ~FALL_BIT;
	}
}

void
sf_label(void)
{
	SL_DISABLE();
	
	sf_nodes [sf_level].f_status |= FALL_BIT;
}

void
sf_return(void)
{
	SL_DISABLE();
	
	sf_nodes [sf_level].f_status &= ~FALL_BIT;
}

void
sf_1switch(void)
{
	flow_node * p;
	
	SL_DISABLE();

	if (sf_level >= MAX_FLOW) {
		warning(es("Statements nested too deeply"); enl());
	}
	else {
		p = & sf_nodes [++sf_level];
		p -> f_type   = K_SWITCH;
		p -> f_status = 0;
	}
}

void
sf_default(void)
{
	SL_DISABLE();
	
	sf_nodes [sf_level].f_status |= DEFAULT_BIT;
}

void
sf_2switch(void)
{
	flow_node * p;
	
	SL_DISABLE();

	p = & sf_nodes [sf_level];
	sf_level--;
	if (	(p -> f_status & DEFAULT_BIT) == 0 ||
		(p -> f_status & BREAK_BIT) ||
		(p -> f_status & FALL_BIT)) {
		sf_nodes [sf_level].f_status |= FALL_BIT;
	}
	else {
		sf_nodes [sf_level].f_status &= ~FALL_BIT;
	}
}

void
sf_1while(void)
{
	flow_node * p;
	
	SL_DISABLE();

	if (sf_level >= MAX_FLOW) {
		warning(es("Statements nested too deeply"); enl());
	}
	else {
		p = & sf_nodes [++sf_level];
		p -> f_type   = K_WHILE;
		p -> f_status = 0;
	}
}

void
sf_2while(void)
{
	SL_DISABLE();
	
	sf_nodes [--sf_level].f_status |= FALL_BIT;	/* Never terminates. */
}

/*
	O u t p u t    R o u t i n e s.
*/
static char * cvt_type(en_tokens tok, bool flag);

/*
	Disable generation of Sherlock macros, but _only_ if we are not
	already generating them!!
	This is done by disabling so_entry(), so_?exit() and so_leave().
*/
void
so_disable(void)
{
	TICK("so_disable");

	if (fn_gen && fn_enable) {
		if (!fn_warn) {
			warning(es("Sherlock macros generated for this function"); enl());
		}
		fn_warn = TRUE;
	}
	else {
		fn_enable = FALSE;
	}
}

/*
	Ouput the cumulative newline count.
*/
static void
so_cnl(void)
{
	int i, count;
	
	SL_DISABLE();
	
	count = cnl_count;
	if (count > 2) {
		count = 2;
	}
	for (i = 0; i < count; i++) {
		io_putc('\n');
	}
	cnl_count = 0;
}

/*
	Output held white space.
*/
static void
so_hws(void)
{
	int i;
	
	SL_DISABLE();

	for (i = 0; i < hws_count; i++) {
		io_putc(hws_buf[i]);
	}
	hws_count = 0;
}

/*
	Output leading white space.
*/
static void
so_lws(void)
{
	int i;
	
	SL_DISABLE();

	for (i = 0; i < lws_count; i++) {
		io_putc(lws_buf[i]);
	}
}

/*
	Just like so_usr except with different names of routines.
*/
static void
so_exxx (void)
{
	FTAG("so_exxx");

	int i = 0;
	int length = 0;
	
	STATB(ftag);

	io_puts("\n"); so_lws(); io_puts("\t");
	length = 16;

	for (i = 0; i < n_param; i++) {

		length += (strlen(pst[i].p_name) + 24);
		if (length > 77) {
			io_puts("\n"); so_lws(); io_puts("\t");
			length = 16 + strlen(pst[i].p_name) + 24;
		}

		switch(pst[i].p_type) {
		case K_CHAR:		io_puts("echar"); break;
		case K_BOOL:		io_puts("ebool"); break;
		case K_UNSIGNED:	io_puts("euint"); break;
		case UNSIGNED_LONG:	io_puts("eulong"); break;
		case K_SIGNED:
		case K_ENUM:
		case K_SHORT:
		case K_INT:			io_puts("eint"); break;
		case K_LONG:		io_puts("elong"); break;
		case STRING_TOK:	io_puts("es"); break;
		case K_STRUCT:
		case K_UNION:
		case LPAREN_TOK:
		case LBRACK_TOK:
		case STAR_TOK:		io_puts("eptr"); break;
		case K_DOUBLE:
		case K_FLOAT:		io_puts("edouble"); break;
		default:
			warning(es("Unknown parameter type: int assumed"); enl());
			io_puts("elong");
		}
		io_puts(open_paren);

		if (pst[i].p_type == K_STRUCT || pst[i].p_type == K_UNION) {
			io_putc('&');
		}
		io_puts(pst[i].p_name);
		io_puts(close_paren);
		if (i < n_param - 1) {
			io_puts("; ecs(); ");
		}
	}

	io_puts("; ecnl()");
	io_puts(close_paren);
	io_puts(";");
	
	STATX(ftag);
}

/*
	Output the entry macro.
*/
void
so_entry(void)
{
	char *p;
	
	SL_DISABLE();

	/* 3/5/89 */
	if (m_flag) {
		warning(es("Entry macro apparently exists"); enl());
	}

	if (!fn_enable || !entry_flag || m_flag || io_in_strm_level != 0) {
		return;
	}

	TRACEPB("so_entry", es(fn_name); enl());

	fn_gen = TRUE;

	if (last_tok == LCURLY_TOK) {
		/* There are no local declarations. */
		io_putc('\n');
		so_lws();
	}
	else if (last_tok == SEMICOLON_TOK) {
		/* There _are_ local declarations. */
		io_putc('\n');
		io_putc('\n');
		so_lws();
	}
	else {
		warning(
			es("so_entry:  can't happen");
			es("last_tok: "); eint(last_tok); enl());

		io_putc('\n'); io_putc('\n');
		so_lws();
	}

	/* Special case the main routine. */
	if (str_eq(fn_name, "main")) {
		so_main();
	}

	if (n_param == 0) {
		if (exxx_flag) {
			io_puts(statb);
		}
		else if (tick_flag) {
			io_puts(tick);
		}
		else {
			io_puts(tickb);
		}
		io_puts(open_paren);
		so_fname();
		io_puts(close_paren);
		io_puts(";");
	}
	else {
		if (tick_flag) {
			io_puts(ptrace);
		}
		else {
			io_puts(pbtrace);
		}
		io_puts(open_paren);
		so_fname();
		io_puts(", ");

		/* Output the params. */
		if (exxx_flag) {
			so_exxx();
		}
		else if (usr_flag) {
			so_usr();
		}
		else if (usr2_flag)
		{
			/* 6 parameters max per sl_fmt statement. */
			int i;
			for (i = 0; i < n_param; i += 6) {
				so_printf(i, i + 6);
			}
		}
		else {
			so_printf(0, n_param);
		}
	}

	/*
		Output exactly one blank line unless the function
		body is empty (next token is '}' or "return".

		Do _not_ output a blank line in front of return so that
		SDEL will work properly.
	*/
	p = hold_nldel();

	if (p == NULL || (  !str_eq(p, "}") && !str_eq(p, "return") )) {
		io_putc('\n');
		io_putc('\n');
	}
	STATX("so_entry");
}

/*
	Output the function name in quotes or the variable named ftag.
*/
static void
so_fname (void)
{
	if (ftag_flag) {
		io_puts("ftag");
	}
	else {
		io_puts(double_quote);
		io_puts(fn_name);
		io_puts(double_quote);
	}
}

/*
	Insert the FTAG or SL_NAME macros at the start of a function.
	This routine is called at the start of body().
*/
void
so_ftag (void)
{
	if (ftag_flag) {
		io_putc('\n');
		so_lws();
		io_puts("FTAG");
		io_puts(open_paren);
		io_puts(double_quote);
		io_puts(fn_name);
		io_puts(double_quote);
		io_puts(close_paren);
		io_puts(";\n");
	}
}

/*
	Insert code to be placed in the main() routine.
*/
static void
so_main (void)
{

#ifdef ENV_DOS
	io_puts(s_init);
	io_puts("();\n");
	so_lws();
	io_puts(s_parse);
	io_puts("(argc, argv, \"++\", \"--\");\n\n");
#endif

#if 0 /* Mac only: Insert the macro in the following macro... */
	
	SL_MAC_INIT(
		SL_VERSION_NAME,	/* version name from sl.h.					*/
		TRUE,				/* Initialize the Macintosh ToolBox.		*/
		TRUE,				/* TRUE: output to Sherlock window.			*/
		"LOG",				/* Name of Sherlock window (if used).		*/
		TRUE,				/* TRUE: open Sherlock window.				*/
		TRUE,				/* TRUE: Insert standard menus.				*/
		SHERLOCK_DEFINED,	/* TRUE: Insert Sherlock menu.				*/
		TRUE,				/* TRUE: Insert first item in apple menu.	*/
		"About...",			/* Title of first item in apple menu.		*/
		0L,					/* CallBack routine for first apple item.	*/
		TRUE,				/* TRUE: Insert menus immediately.			*/
		0L,					/* Event callback routine.					*/
		0L					/* User dump callback routine.				*/
	);
	
	SL_DIALOG("++", "--", "Log File...", "log");
	
#endif /* end comment */

#if 0 // This is out of date: the SL_MAC_INIT macro does not even exist!
	
	io_puts(s_init);
	io_puts("(\n");
	io_puts("\t\tSL_VERSION_NAME,\t/* version name from sl.h.\t\t\t\t\t*/\n");
	io_puts("\t\tTRUE,\t\t\t\t/* Initialize the Macintosh ToolBox.\t\t*/\n");
	io_puts("\t\tTRUE,\t\t\t\t/* TRUE: output to Sherlock window.\t\t\t*/\n");
	io_puts("\t\t\"LOG\",\t\t\t\t/* Name of Sherlock window (if used).\t\t*/\n");
	io_puts("\t\tTRUE,\t\t\t\t/* TRUE: open Sherlock window.\t\t\t\t*/\n");
	io_puts("\t\tTRUE,\t\t\t\t/* TRUE: Insert standard menus.\t\t\t\t*/\n");
	io_puts("\t\tSHERLOCK_DEFINED,\t/* TRUE: Insert Sherlock menu.\t\t\t\t*/\n");
	io_puts("\t\tTRUE,\t\t\t\t/* TRUE: Insert first item in apple menu.\t*/\n");
	io_puts("\t\t\"About...\",\t\t\t/* Title of first item in apple menu.\t\t*/\n");
	io_puts("\t\t0L,\t\t\t\t\t/* CallBack routine for first apple item.\t*/\n");
	io_puts("\t\tTRUE,\t\t\t\t/* TRUE: Insert menus immediately.\t\t\t*/\n");
	io_puts("\t\t0L,\t\t\t\t\t/* Event callback routine.\t\t\t\t\t*/\n");
	io_puts("\t\t0L\t\t\t\t\t/* User dump callback routine.\t\t\t\t*/\n");
	io_puts("\t);\n\n");
	
	io_puts("\tSL_DIALOG(\"++\", \"--\", \"Log File...\", \"log\");\n\n");
		
#endif

	so_lws();
}

/*
	Output the parameters from start to end.
	Use sl_fmt if usr2_flag is TRUE, otherwise printf.
*/
static void
so_printf(reg int start, reg int end)
{
	bool flag;
	reg int i;
	reg int length;
	reg int newlength;
	reg char *p;
	
	SL_DISABLE();
	
	if (end > n_param) {
		end = n_param;
	}

	flag = FALSE;
	length = 32 + strlen(fn_name);

	/* Output the printf format specification. */
	if (usr2_flag) {
		char buffer [10];
		io_puts("sl_fmt");
		cvt_int	(buffer, 10, end - start);
		io_puts(open_paren);
		io_puts(double_quote);
	}
	else {
		io_puts("printf");
		io_puts(open_paren);
		io_puts(double_quote);
	}
	
	for (i = start; i < end; i++) {
		p = cvt_type(pst[i].p_type, TRUE);
		if (p != NULL) {

			/* Make sure the line would not be too long. */
			length += strlen(p) + 2;
			if (length >= 77 && i < n_param-1) {
				io_puts("\\\n");
				length = strlen(p) + 2;
				flag = TRUE;
			}

			/* Output the format specification. */
			io_puts(p);
			if (i < n_param - 1) {
				io_puts(", ");
			}
		}
	}

	/* End the printf format string. */
	io_puts("\\n\", ");
	length += 6;

	/* Go to the next line if we can't get _all_ the params on it. */
	for (newlength = length, i = start; i < end; i++) {
		newlength += strlen(pst[i].p_name) + 2;
		if (usr2_flag) {
			 if (pst[i].p_type != STRING_TOK) {
				newlength += strlen(usr2_cvt(pst[i].p_type)) + 2;
			}
		}
		else {
			if (pst[i].p_type == K_BOOL) {
				newlength += strlen("sl_sbout") + 2;
			}
		}
	}

	if (newlength >= 77 || flag) {
		io_puts("\n"); so_lws(); io_puts("\t");
		length = 16;
	};

	/* Output the list of arguments. */
	for (i = start; i < end; i++) {
		p = cvt_type(pst[i].p_type, FALSE);
		if (p != NULL) {

			/* Compute the new length of the line. */
			length += strlen(pst[i].p_name) + 2;

			if (	pst[i].p_type == K_STRUCT ||
					pst[i].p_type == K_UNION
				) {
				length++;
			}
			
			if (usr2_flag) {
				if (pst[i].p_type != STRING_TOK) {
					length += strlen(usr2_cvt(pst[i].p_type)) + 2;
				}
			}
			else {
				if (pst[i].p_type == K_BOOL) {
					length += strlen("sl_sbout") + 2;
				}
			}

			/* Go to a new line if it would be too long. */
			if (length >= 77 && i < n_param - 1) {
				io_puts("\n"); so_lws(); io_puts("\t");
				length = 16;
			}

			if (usr2_flag) {
				if (pst[i].p_type == STRING_TOK) {
					io_puts(pst[i].p_name);
				}
				else if (
					pst[i].p_type == K_STRUCT ||
					pst[i].p_type == K_UNION
				) {
					io_puts(usr2_cvt(pst[i].p_type));
					io_puts(open_paren);
					io_puts("&");
					io_puts(pst[i].p_name);
					io_puts(close_paren);
				}
				else {
					io_puts(usr2_cvt(pst[i].p_type));
					io_puts(open_paren);
					io_puts(pst[i].p_name);
					io_puts(close_paren);
				}
			}
			else {
				if (pst[i].p_type == K_BOOL) {
					io_puts("cvt_bool");
					io_puts(open_paren);
					io_puts(pst[i].p_name);
					io_puts(close_paren);
				}
				else if (
					pst[i].p_type == K_STRUCT ||
					pst[i].p_type == K_UNION
				) {
					io_putc('&');
					io_puts(pst[i].p_name);
				}
				else {
					io_puts(pst[i].p_name);
				}
			}
			
			if (i < end - 1) {
				io_puts(", ");
			}
		}
	}
	io_puts(close_paren);
	io_puts(close_paren);
	io_puts(";");
}

static bool old_m;
static bool m_warn;

void
so_1exit(void)
{
	FTAG("so_1exit");

	if (!fn_enable) {
		return;
	}

	old_m = m_flag;
	m_warn = FALSE;

	if (m_flag) {
		warning(es("Return inside macro expansion"));
		m_warn = TRUE;
	}

	if (!exit_flag || m_flag || old_m || io_in_strm_level != 0) {
		return;
	}

	TRACEPB(ftag, es(fn_name); enl());

	fn_gen = TRUE;

	/* Kill the hold buffer. */
	hold_kill();

	/* Output appropriate white space. */
	if (first_tok) {
		so_cnl();
		so_lws();
	}
	else {
		so_hws();
	}

	STATX("so_1exit");
}

/*
	ret_exp_flag is TRUE if an expression exists following a return.

	When exxx_flag, return_flag and ret_exp_flag are all TRUE,
	we would like to generate:
	
		TRACEPX(ftag, eret(); exxx(rval); enl());
		return rval;
		
	where rval the expression following the 'return' statement.
	However, at present we just generate:
	
		STATX(ftag);
		return rval;
		
	because there is no way to duplicate rval.
*/
void
so_2exit(bool ret_exp_flag)
{
	FTAG("so_2exit");

	/* Don't worry about current macro expansions, only old ones. */
	if (!fn_enable || !exit_flag || old_m || io_in_strm_level != 0) {
		return;
	}

	TRACEPB(ftag, es(fn_name); ecs(); ebool(ret_exp_flag); enl());

	fn_gen = TRUE;

	if (!ret_exp_flag && fn_type != K_VOID) {
		warning(es("Void return from function of non-void type"));
	}

	if (tick_flag) {
		/* Generate just a return. */
		io_puts("return");
	}
	else if (
		return_flag && exxx_flag &&
		fn_type != K_STRUCT && fn_type != K_UNION &&
		fn_type != K_VOID
	) {
		so_tracepx1();
	}
	else if (
		return_flag && !exxx_flag &&
		fn_type != K_STRUCT && fn_type != K_UNION
	) {
		/* Generate RETURN_XXX(ftag, rval); */
		so_return();
	}
	else {
		/* Generate STATX(ftag); return. */
		io_puts(statx);
		io_puts(open_paren);
		so_fname();
		io_puts(close_paren);
		io_puts(";\n");
		so_lws();
		io_puts("return");
	}

	TICKX("so_2exit");
}

void
so_3exit(void)
{
	FTAG("so_3exit");
	TICKB(ftag);

	/* Don't worry about current macro expansions. */
	if (
		fn_enable && exit_flag && return_flag &&
		!old_m && io_in_strm_level == 0 &&
		fn_type != K_STRUCT && fn_type != K_UNION
	) {
		fn_gen = TRUE;
	
		if (exxx_flag && fn_type != K_VOID) {
			so_tracepx2();
		}
		else if (!tick_flag && !exxx_flag) {
			io_puts(close_paren);
		}
	}

	STATX(ftag);
}

void
so_leave(void)
{
	FTAG("so_leave");

	/* Bug fix: 3/5/89 */
	if (!fn_enable || tick_flag || !exit_flag || m_flag || io_in_strm_level != 0) {
		return;
	}
	
	TRACEPB(ftag, es(fn_name); enl());

	fn_gen = TRUE;
	
	/* Examine flow control to see if we need anything. */
	if ((sf_nodes [0].f_status & FALL_BIT) == 0) {
		STATX("so_leave");
		return;
	}

	/* Kill the buffer, containing white space and }. */
	hold_kill();

	/* Output one blank line and leading white space. */
	io_putc('\n');
	io_putc('\n');
	so_lws();

	/* Output STATX macro */
	io_puts(statx);
	io_puts(open_paren);
	so_fname();
	io_puts(close_paren);
	io_puts(";");

	/* One trailing newline. */
	io_putc('\n');

	/* Output the teminating brace of the function. */
	io_putc('}');

	STATX("so_leave");
}

static void
so_return (void)
{
	FTAG("so_return");
	TRACEPB(ftag, es_op(fn_type); enl());
	ASSERT(fn_type != K_STRUCT && fn_type != K_UNION);

	switch(fn_type) {

	case K_VOID:	 	io_puts(r_void);	break;
	case K_CHAR:	 	io_puts(r_char);	break;
	case K_BOOL:	 	io_puts(r_bool);	break;
	case K_UNSIGNED: 	io_puts(r_uint);	break;
	case UNSIGNED_LONG:	io_puts(r_ulong);	break;
	case K_ENUM:
	case K_SIGNED:
	case K_SHORT:
	case K_INT:	 	 io_puts(r_int);	break;
	case K_LONG:	 io_puts(r_long);	break;
	case STRING_TOK: io_puts(r_string);	break;
	case LPAREN_TOK:
	case LBRACK_TOK:
	case STAR_TOK:	 io_puts(r_ptr);	break;
	case K_FLOAT:	 io_puts(r_float);	break;
	case K_DOUBLE:	 io_puts(r_double);	break;

	default:		
		warning(es("Unknown expression type--no RETURN macro generated"); enl());
		io_puts("return");
		break;
	}

	/* Output function name. */
	io_puts(open_paren);
	so_fname();
		
	/*
		Output a comma for all types except void, struct or union.
		No white space is needed because the white space after
		the original "return" will be used.
	*/
	if (fn_type != K_VOID) {
		io_putc(',');
	}

	TICKX(ftag);
}

static void
so_tracepx1 (void)
{
	FTAG("so_tracepx1");
	TRACEPB(ftag, es_op(fn_type); enl());
	ASSERT(fn_type != K_STRUCT && fn_type != K_UNION && fn_type != K_VOID);
	
	/* Output "TRACEX(ftag, " */
	io_puts(pxtrace);
	io_puts(open_paren);
	so_fname();
	io_puts(", eret(); ");
	
	/* Output the name of the exxx routine. */
	switch(fn_type) {
	
	case K_CHAR:	 	io_puts("echar");	break;
	case K_BOOL:	 	io_puts("ebool");	break;
	case K_UNSIGNED: 	io_puts("euint");	break;
	case UNSIGNED_LONG:	io_puts("eulong");	break;
	case K_ENUM:
	case K_SIGNED:
	case K_SHORT:
	case K_INT:	 	 	io_puts("eint");	break;
	case K_LONG:	 	io_puts("elong");	break;
	case STRING_TOK: 	io_puts("es");		break;
	case LPAREN_TOK:
	case LBRACK_TOK:
	case STAR_TOK:	 	io_puts("eptr");	break;
	case K_FLOAT:
	case K_DOUBLE:	 	io_puts("edouble");	break;

	default:		
		warning(es("Unknown expression type--no exxx call generated"); enl());
		break;
	}
	
	/* Output the open parenthesis before the return expression. */
	io_puts(open_paren);
	
	STATX(ftag);
}

static void
so_tracepx2 (void)
{
	FTAG("so_tracepx2");
	short i = 0;
	TRACEPB(ftag, es_op(fn_type); enl());
	ASSERT(fn_type != K_STRUCT && fn_type != K_UNION && fn_type != K_VOID);
	
	/*
		Output the parenthesis following the return expression
		which has already been copied.
	*/
	io_puts(close_paren);
	
	/* End the TRACEPX statement. */
	io_puts("; enl()");
	io_puts(close_paren);
	io_puts(";");
	
	/* Start the return statement. */
	io_puts("\n");
	so_lws();
	io_puts("return ");
	
	/*
		Copy the return expression.
		The copy buffer ends with a semicolon: don't print it.
		
		WARNING: this computes the expression twice!
		If this is a problem, return the value a local variable!
	*/
	--copy_count;	/* Point copy_count-1 at the last character. */
	if (copy_count - 1 >= 0 && copy_buf[copy_count - 1] == ';') {
		--copy_count;
	}

	for (i = 0; i < copy_count; i++)
		io_putc(copy_buf[i]);

	STATX(ftag);
}

static void
so_usr(void)
{
	int i;
	int length;
	
	SL_DISABLE();

	io_puts(" sl_lpout();\n");
	so_lws(); io_puts("\t");
	length = 16;

	for (i = 0; i < n_param; i++) {

		length += (strlen(pst[i].p_name) + 24);
		if (length > 77) {
			io_puts("\n"); so_lws(); io_puts("\t");
			length = 16 + strlen(pst[i].p_name) + 24;
		}

		switch(pst[i].p_type) {
		case K_CHAR:		io_puts("sl_cout"); break;
		case K_BOOL:		io_puts("sl_bout"); break;
		case K_UNSIGNED:	io_puts("sl_uiout"); break;
		case UNSIGNED_LONG:	io_puts("sl_ulout"); break;
		case K_SIGNED:
		case K_ENUM:
		case K_SHORT:
		case K_INT:			io_puts("sl_iout"); break;
		case K_LONG:		io_puts("sl_lout"); break;
		case STRING_TOK:	io_puts("sl_sout"); break;
		case K_STRUCT:
		case K_UNION:
		case LPAREN_TOK:
		case LBRACK_TOK:
		case STAR_TOK:		io_puts("sl_pout"); break;
		case K_DOUBLE:
		case K_FLOAT:		io_puts("sl_fout"); break;
		default:
			warning(es("Unknown parameter type: int assumed"); enl());
			io_puts("sl_iout");
		}
		io_puts(open_paren);

		if (pst[i].p_type == K_STRUCT || pst[i].p_type == K_UNION) {
			io_putc('&');
		}
		io_puts(pst[i].p_name);
		io_puts(close_paren);
		if (i < n_param - 1) {
			io_puts("; sl_csout(); ");
		}
	}

	io_puts("; sl_rpout()");
	io_puts(close_paren);
	io_puts(";");
}

static char *
cvt_type(en_tokens type, bool flag)
{
	SL_DISABLE();
	
	switch(type) {
	case K_CHAR:		return "%c";
	case K_UNSIGNED:	return "%u";
	case UNSIGNED_LONG:	return "%lu";
	case K_BOOL:		return "%s";
	case K_ENUM:
	case K_SIGNED:
	case K_SHORT:
	case K_INT:			return "%d";
	case K_LONG:		return "%ld";
	case STRING_TOK:	return "%s";
	case K_STRUCT:
	case K_UNION:
	case LBRACK_TOK:
	case LPAREN_TOK:
	case STAR_TOK:		return "%p";
	case K_DOUBLE:
	case K_FLOAT:		return "%g";
	default:
		if (flag) {
			warning(
				es("Unknown formal type: "); es_op(type);
				es(", int assumed"); enl());
		}
		return "%d";
	}
}

static char *
usr2_cvt (en_tokens type)
{
	SL_DISABLE();
	
	switch(type) {
	case K_CHAR:		return "cvt_char";
	case K_UNSIGNED:	return "cvt_uint";
	case UNSIGNED_LONG:	return "cvt_ulong";
	case K_BOOL:		return "cvt_bool";
	case K_ENUM:
	case K_SIGNED:
	case K_SHORT:
	case K_INT:			return "cvt_int";
	case K_LONG:		return "cvt_long";
	case K_STRUCT:
	case K_UNION:
	case LBRACK_TOK:
	case LPAREN_TOK:
	case STAR_TOK:		return "cvt_ptr";
	case K_DOUBLE:
	case K_FLOAT:		return "cvt_double";
	default:
		warning(
			es("Unknown formal type: "); es_op(type);
			es(", int assumed"); enl());

		return "cvt_int";
	}
}

static void
sd_dump()
{
	SL_DISABLE();

	ecnls(2);
	es("cur_typedef: "); ebool(cur_typedef); enl();
	es("cur_hptype:  "); es_op(cur_hptype); enl();
	es("cur_hftype:  "); es_op(cur_hftype); enl();
	es("cur_ptype:   "); es_op(cur_ptype); enl();
	es("cur_ftype:   "); es_op(cur_ftype); enl();
	es("cur_pid:     "); es(cur_pid); enl();
	es("cur_fid:     "); es(cur_fid); enl();
	es("fn_type:     "); es_op(fn_type); enl();
	es("fn_name:     "); es(fn_name); enl();
	ecnls(2);
}
