/*
	SPP: symbol table routines.

	source:  SPPst.c
	started: November 4, 1985
	version: March 16, 1996.
*/

#include "SPP.h"

#include "SPPpr.h"
#include "SPPst.h"

#include "SPPerr.h"	/* Library additions. */
#include "SPPlst.h"
#include "SPPobj.h"
#include "SPPstr.h"

/* Define static routines. */
static int st_hash(char *s);

/*
	Define the format of symbol table nodes.
*/
#define ST_PRIME 101

typedef struct st_node_struct st_node;

static struct st_node_struct {
	LIST2;
	char *		st_name;
	en_tokens	st_type;	/* type for printf	*/
	bool		st_tdflag;	/* typedef flag		*/
};

static st_node * st_ht [ST_PRIME];	/* The hash table. */

/*
	Define the list of symbol tables and the list of nodes.
*/
typedef struct st_level_struct st_level;
typedef struct st_list_struct  st_list;

static struct st_level_struct {
	LIST;
	st_list *	st_lev_cur_list;
};

static st_level * level_list;

static struct st_list_struct {
	LIST;
	st_node *	st_theNode;
};

static int cur_level = 0;
static st_list * cur_list = NULL;

/*
	Enter a new level of the symbol table.
*/
void
st_begin(void)
{
	st_level * lp;

	TRACEPB("st_begin",
		es("cur_level = "); eint(cur_level+1); enl());

	/* Remember the old cur_list. */
	lp = obj_new(st_level, OBJ_ST_LEVEL_STRUCT);
	lp -> st_lev_cur_list = cur_list;
	lst_add(lp, &level_list);
	
	/* Start over with a new list. */
	cur_list = NULL;
	cur_level++;
	
	STATX("st_begin");
}

/*
	Exit a level of the symbol table.
*/
void
st_end(void)
{
	reg st_list *	cp;
	reg st_level *	lp = level_list;
	reg int			i;

	TICK("st_end");

	if (cur_level <= 0) {
		fatal(es("st_end: cur_level <= 0"));
	}
	
	ASSERT(lp);
	
	/* Unlink all nodes from the bucket list. */
	lst_every(st_list, cp, &cur_list) {
		lst2_remove(cp -> st_theNode);
		obj_free(cp -> st_theNode -> st_name);
		obj_free(cp -> st_theNode);
	}
	
	/* Free all nodes. */
	lst_free_all(&cur_list);
	
	/* Enter the new level. */
	cur_list = lp -> st_lev_cur_list;
	cur_level--;
	
	/* Get rid of the old level. */
	lst_remove(lp, &level_list);
	obj_free(lp);
	
	if (cur_level == 0) {
		/* Clean out the list2 headers. */
		for(i = 0; i < ST_PRIME; i++) {
			lst2_free_all(&st_ht[i]);
		}
	}
}

/*
	Place a symbol at the start of the hash list.
	This will insure that any new definition hides previous defs.
*/
#undef symbol /* Symbol is defined in the Mac world to be a font. */

void
st_enter (reg char *symbol, reg en_tokens type, bool td_flag)
{
	FTAG("st_enter");
	
	reg st_node *	p;
	reg st_list *	lp;
	reg int 		hash = st_hash(symbol);

	TRACEPB(ftag,
		es(symbol); ecs(); es_op(type); ecs(); ebool(td_flag); enl());

	if (symbol == NULL) {
		error(es("st_enter: internal: NULL symbol"));
		symbol = "";
		return;
	}

	/* Create a new symbol table node. */
	p = obj_new(st_node, OBJ_ST_NODE_STRUCT);
	p -> st_type	= type;
	p -> st_name    = str_alloc(symbol, OBJ_TYPDEF_STRING);
	p -> st_tdflag  = td_flag;
	
	/* Hang the node from hash table. */
	lst2_add(p, &st_ht[hash]);
	
	/* Create a new list node. */
	lp = obj_new(st_list, OBJ_ST_LIST_STRUCT);
	lp -> st_theNode = p;
	
	/* Link the node to the level list. */
	lst_add(lp, &cur_list);

	STATX("st_enter");
}

/*
	Compute the hash function for a local symbol.
*/
static int
st_hash (reg char * sym)
{
	reg int hash = 0;

	SL_DISABLE();
	
	while(*sym) {
		hash *= 3;
		hash += (int) *sym++;
		hash %= ST_PRIME;
	}
	return hash;
}

/*
	Initialize the symbol tables for the very first time.
*/
void
st_init(void)
{
	reg int		i;

	TICK("st_init");

	/* Initialize the hash table. */
	for(i = 0; i < ST_PRIME; i++) {
		st_ht[i] = NULL;
	}
	
	/* Bug fix: 3/16/90 */
	st_begin();
}

/*
	Look up a symbol in the local symbol table.
	Return a type token.
*/
bool
st_lookup (reg char * symbol, en_tokens * type, bool * td_flag)
{
	FTAG("st_lookup");

	reg st_node *	np;
	reg int			hash = st_hash(symbol);
	
	lst2_every(st_node, np, &st_ht[hash]) {
		if(str_eq(symbol, np -> st_name)) {

			TRACEP(ftag,
				es(symbol); ecs(); es_op(np -> st_type); ecs();
				ebool(np -> st_tdflag); enl());

			*type    = np -> st_type;
			*td_flag = np -> st_tdflag;
			return TRUE;
		}
	}

	TRACEP(ftag, es(symbol); eret(); ebool(FALSE); enl());
	return FALSE;
}
