/*
	SPP: Token parsing routines.
	
	source:  SPPtok.c
	started: October 7, 1985
	version: March 16, 1996.
*/

#include "SPP.h"

#include "SPParg.h"
#include "SPPhold.h"
#include "SPPmain.h"
#include "SPPpr.h"
#include "SPPnext.h"
#include "SPPtok.h"

#include "SPPenv.h"	/* Library additions. */
#include "SPPerr.h"
#include "SPPstr.h"

/*
	Globals describing tokens.

	The code assumes that no token or string will ever be longer than
	MAX_SYMBOL.  If that ever is not so the program may crash.  Thus,
	MAX_SYMBOL should be very large -- say 1000 or more.
*/
char		ch = '@';				/* The next character of input.			*/
en_tokens	token = NULL_TOK;		/* The current token. 					*/

int			t_subtype;				/* The subtype of token class.			*/
char		t_symbol [MAX_SYMBOL];	/* The spelling of the token.			*/
int			t_length;				/* The length of the token (in the text)*/
long		t_value;				/* Value of integer constants.			*/

int			last_tok  = NULL_TOK;	/* Previous token.		*/
int			last_ch   = '@';		/* Last char.			*/
bool		first_tok = FALSE;		/* True if token is first in line.	*/
bool		con_flag = FALSE;		/* True if in PP constant expr. 	*/

/*
	Declare internal variables and routines.
*/
static int	 is_base_digit	(int);
static void	 scan_number	(int);
static en_tokens t_int		(void);

/*
	Return >= 0 if ch is a digit in the indicated base.
	Otherwise, return -1.
*/
static int
is_base_digit(int base)
{
	TRACECH("is_base_digit");

	if (ch >= '0' && ch <= '9') {
		STATX("is_base_digit");
		return ch - '0';
	}
	else if (base != 16) {
		STATX("is_base_digit");
		return -1;
	}
	else if (ch >= 'a' && ch <= 'f') {
		STATX("is_base_digit");
		return ch - 'a' + 10;
	}
	else if (ch >= 'A' && ch <= 'F') {
 		STATX("is_base_digit");
 		return ch - 'A' + 10;
	}
	else {
		STATX("is_base_digit");
		return -1;
	}
}

/*
	Get value of a string of digits into t_value.
	Continue until a non base digit is found.
*/
static void
scan_number(int base)
{
	int result;

	TRACECH("scan_number");

	t_value = 0;
	while (	(result = is_base_digit(base)) != -1) {
		t_value = ((long)base * t_value) + (long) result;
		io_next();
	}

	TICKX("scan_number");
}

/*
	Skip blanks and tabs,  but not newlines.
*/
void
skip_bl(void)
{
	TICKB("skip_bl");

	while (ch == ' ' || ch == '\t' || ch == '\v' || ch == '\f') {
		io_next();
	}
	
	STATX("skip_bl");
}

/*
	Scan past a comment. Allow nested comments if nest_flag is TRUE.
	Surprisingly, this routine needs to be as fast as possible.
*/
void
skip_comment(void)
{
	reg int clevel;
	reg int start_line;

	TRACECH("skip_comment");

	/* Save starting line number for run-on comments. */
	start_line = err_line;
	clevel = 1;

	/* Do not buffer the comment. */
	hold_flush();
	hold_flag = FALSE;

	for (;;) {
		switch (ch) {

		case END_FILE:
			error(
				es("File ends in a comment starting at line ");
				elong(start_line); enl());

			hold_flag = TRUE;

			STATX("skip_comment");
			return;

		case '\n':
			/* Keep track of line numbering. */
			io_next();

			/*
				Bump the line count and output a newline
				if this is used in a standalone preprocessor.
				Do NOT allow directives here.
			*/
			bump_line();
			/* Bug fix: 6/26/89 do NOT call begin_line!! */
			/* --- begin_line(FALSE); --- */
			first_tok = TRUE;
			continue;

		case '/':

			io_next();
			if (ch == '*') {
				io_next();
				if (nest_flag) {
					clevel++;
				}
			}
			continue;

		case '*':

			io_next();
			if (ch == '/') {
				io_next();
				if (--clevel == 0) {
					hold_flag = TRUE;
					STATX("skip_comment");
					return;
				}
			}
			continue;


		default:
			io_next();
		}
	}
}

/*
	Skip characters up to but NOT including a newline.
*/
void
skip_1line(void)
{
	TICKB("skip_1line");

	while (ch != END_FILE && ch != '\n') {
		io_next();
	}

	STATX("skip_1line");
}

/*
	Skip characters up to and including a newline.
*/
void
skip_past(void)
{
	FTAG("skip_past");
	TRACEPB(ftag, es("line "); elong(err_line); enl())

	while (ch != END_FILE && ch != '\n') {
		io_next();
	}

	if (ch == '\n') {
		io_next();
		bump_line();
	}

	STATX("skip_past");
}

/*
	Skip to the end of the current preprocessor directive.
	I.e., skip to the first newline not contained in a comment.
	
	It is safe to call this routine even if all the line has _already_
	been skipped, that is, if the current character is a newline.
*/
void
skip_pp(void)
{
	TICKB("skip_pp");

	while (ch != '\n' && ch != END_FILE) {
		if (ch == '/') {
			/* Possible comment. */
			io_next();
			if (ch == '*') {
				io_next();
				skip_comment();
			}
			else if (slc_flag && ch == '/') {
				/* 8/3/89 Single-line comment */
				while (ch != END_FILE && ch != '\n') {
					io_next();
				}
				STATX("skip_pp");
				return;
			}
			else {
				io_pushback(ch);
				ch = '/';
				hold_flush();
				STATX("skip_pp");
				return;
			}
		}
		else {
			io_next();
		}
	}
	hold_flush();
	STATX("skip_pp");
}

/*
	Skip blanks, tabs, and comments.
	Also skip newlines if nl_flag is TRUE;
*/
void
skip_ws(bool nl_flag)
{
	TRACEPB("skip_ws", ebool(nl_flag); enl());

	for(;;) {

		switch(ch) {

		case ' ':
		case '\t':
		case '\f':
		case '\v':
			io_next();
			continue;

		case '\n':
			if (nl_flag) {
				bump_line();
				io_next();
				continue;
			}
			else {
				/* 2/16/89 */
				goto done;
			}

		case '/':
			io_next();
			if (ch == '*') {
				io_next();
				skip_comment();
				continue;
			}
			else if (slc_flag && ch == '/') {
				/* 8/3/89 Single-line comment */
				while (ch != END_FILE && ch != '\n') {
					io_next();
				}
				continue;
			}
			else {
				io_pushback(ch);
				ch = '/';
				goto done;
			}

		default:
			goto done;
		}
	}
done:
	STATX("skip_ws");
}

/*
	Copy an identifier into symbol[] and its length in the global t_length.
	Surprisingly, this routine should be as fast as possible.
*/
#undef symbol /* symbol is defined in the Mac world to be a font! */

void
t_id(reg char * symbol, reg int	max_length)
{
	FTAG("t_id");

	reg int length = 0;

	TRACEPB(ftag, eptr(symbol); ecs(); eint(max_length); enl());

	max_length--;
	while (isid2(ch) && length < max_length) {
		*symbol++ = ch;
		length++;
		io_next();
	}
	*symbol = '\0';
	t_length = length;

	if (length >= max_length) {
		error(es("Identifier too long"));
	}

	TICKX("t_id");
}

/*
	Parse an integer constant (octal, decimal or hexadecimal) or float.
	Put the value in t_value if it is an integer.

	dot_flag is TRUE if a decimal point has been seen.

	Return the token type (INT_TOK, LONG_TOK, FLOAT_TOK).

	Legal integer forms:	ddd,	0ddd,	0xddd
	Legal float forms:	xxx.yyyE+zzz

	+-zzz is optional
	one of xxx and yyy may be omitted
	one of . and E may be omitted
*/
en_tokens
t_number(bool dot_flag)
{
	en_tokens result;
	bool need_exp = FALSE;

	TRACECH("t_number");

	/* Defaults. */
	t_value = 0;

	if (dot_flag) {
		goto frac_part;
	}

	/* Integer part. */

	result = t_int();
	if (ch == '.') {
		io_next();
		goto frac_part;
	}
	else if (ch == 'e' || ch == 'E') {
		goto exp_part;
	}
	else {
		STATX("t_number");
		return result;
	}

	/* Fraction part. */
frac_part:	
	t_int();

exp_part:
	if (ch == 'e' || ch == 'E') {
		need_exp = TRUE;
		io_next();
	}
	if (ch == '+' || ch == '-') {
		need_exp = TRUE;
		io_next();
	}
	if (ch >= '0' && ch <= '9') {
		t_int();
	}
	else if (need_exp) {
		error(es("Ill formed floating constant"));
	}

	if (ch == 'l' || ch == 'L' || ch == 'f' || ch == 'F') {
		io_next();
	}

	STATX("t_number");
	return FLOAT_TOK;
}

static en_tokens
t_int(void)
{
	TRACECH("t_int");

	/* Leading 0 or 0x changes base. */	
	if (ch == '0') {
		io_next();
		if (ch == 'x' || ch == 'X') {
			io_next();
			scan_number(16);
		}
		else if (isdigit(ch)) {
			scan_number(8);
		}
		else {
			/* Lone '0'. */
			t_value = 0;
		}
	}
	else {
		scan_number(10);
	}

	if (ch == 'l' || ch == 'L') {
		io_next();
		if(ch == 'u' || ch == 'U') {
			io_next();
		}
		STATX("t_int");
		return LONG_TOK;
	}
	else if (ch == 'u' || ch == 'U') {
		io_next();
		if (ch == 'l' || ch == 'L') {
			STATX("t_int");
			return LONG_TOK;
		}
		else {
			STATX("t_int");
			return INT_TOK;
		}
	}
	else {
		STATX("t_int");
		return INT_TOK;
	}
}

/*
	Copy a string into out[] and puts its length in the global t_length.
	Copy the opening or closing delimiters if the copy_flag is TRUE.

	This is used to parse both strings and character constants.
*/
void
t_string(
	reg char *	out,		/* The output buffer.			*/
	reg int		max_out,	/* The size of out[].			*/
	bool		copy_flag)	/* TRUE: Copy the delimiters	*/
{
	FTAG("t_string");

	reg int	length;
	int		start_line;
	char *	start_string;
	int	delim;

	TRACECH(ftag);

	/* Save starting line number for error messages. */
	start_line = err_line;
	start_string = out;

	/* Handle the opening single or double quote */
	delim = ch;
	io_next();
	length = 0;
	if (copy_flag) {
		*out++ = delim;
		length++;
	}

	max_out--;
	while (length < max_out) {

		switch(ch) {

		case END_FILE:
		case '\n':
			goto runon1;

		case '"':
		case '\'':
			if (ch == delim) {
				io_next();
				if(copy_flag) {
					*out++ = delim;
					length++;
				}
				*out++ = '\0';
				t_length  = length;
				TRACEPX(ftag,es("<"); es(start_string); es(">\n"));
				return;
			}
			else{
				*out++ = ch;
				length++;
				io_next();
			}
			continue;

		case '\\':

			io_next();
			if (ch == END_FILE) {
				goto runon1;
			}

#ifdef BACKSLASH_NEWLINE
			else if (ch == '\n') {
				/* Ignore back slash and newline. */
				err_line++;
				io_next();
			}
#endif

			else {
				*out++ = '\\';
				*out++ = ch;
				length += 2;
				io_next();
			}
			continue;

		default:
			*out++ = ch;
			length++;
			io_next();
		}
	}

	error(
		es("String starting at line "); eint(start_line);
		es(" is too long"); enl());

	*out = '\0';
	t_length = length;
	STATX("t_string");
	return;

runon1:
	*out   = '\0';
	error(es("String crosses a line: <"); es(start_string); es(">"); enl());
	
	t_length  = length;
	STATX("t_string");
}

/*
	Parse a string (including delimiters) from in[] to out.
	Return the length of the ORIGINAL string.
*/
int
in_string(
	reg char *	in,			/* The output buffer	*/
	reg char *	out,		/* The output buffer.	*/
	reg int		max_out)	/* The size of out[].	*/
{
	FTAG("in_string");

	reg int	length;
	int		start_line;
	char *	start_string;
	int		delim;

	TRACEPB(ftag,
		es("<"); es(in); es(">, ");
		eptr(out); ecs(); eint(max_out); enl());

	/* Save starting line number for error messages. */
	start_line = err_line;
	start_string = out;

	/* Copy the opening single or double quote */
	*out++ = delim = *in++;
	length = 1;
	max_out--;
	while (length < max_out) {

		switch(*in) {

		case END_FILE:
		case '\n':
			goto runon1;

		case '"':
		case '\'':
			if (*in == delim) {
				*out++ = *in++;
				*out++ = '\0';
				length++;
				hold_flag = TRUE;

				STATX("in_string");
				return length;
			}
			else{
				*out++ = *in++;
				length++;
			}
			continue;

		case '\\':

			in++;
			if (*in == END_FILE) {
				goto runon1;
			}

#ifdef BACKSLASH_NEWLINE
			else if (*in == '\n') {
				/* Ignore back slash and newline. */
				err_line++;
				in++;

				/* Keep track of ORIGINAL length. */
				length += 2;
			}
#endif

			else {
				*out++ = '\\';
				*out++ = *in++;
				length += 2;
			}
			continue;

		default:
			*out++ = *in++;
			length++;
		}
	}

	error(
		es("Stringized argument starting at line ");
		eint(start_line); es(" is too long"); enl());

	*out = '\0';
	hold_flag = TRUE;

	STATX("in_string");
	return length;

runon1:
	*out   = '\0';
	error(
		es("Stringized argument crosses a line: <");
		es(start_string); es(">"); enl());

	hold_flag = TRUE;

	STATX("in_string");
	return length;
}

/*
	Reinitialize this module.
*/
void
tok_init_again(void)
{

	ch			= '@';		/* The next character of input.		*/
	token		= NULL_TOK;	/* The current token. 				*/
	last_tok	= NULL_TOK;	/* Previous token.					*/
	last_ch		= '@';		/* Last char.						*/
	first_tok	= FALSE;	/* True if token is first in line.	*/
	con_flag	= FALSE;	/* True if in PP constant expr. 	*/
}
