/*
	SPP: Application-specific io/library
		
	source:  SPPio.c
	started: December 1, 1990
	version: March 16, 1996.
	
	 To do:  err_line is long:  io_file_line is int.
*/

/*
	There are some *subtle* points concerning Sherlock lurking in this code:
	
	In effect, Sherlock macros act like "software interrupts".
	Any routine using Sherlock macros can generate a call to
	end_quit via the event "call back" routine in init.c.
		
	Keeping track of the proper state of the open files would require
	locking out Sherlock tracing calls during "io critical sections,"
	i.e., sections in which io_open_files was being updated.
		
	Rather than do this, we will simply not worry about whether the
	io_open_files list is correct when end_quit is called and we will just
	use the standard library exit function to close all open files.
		
	If quit is called while files are being opened or closed, io_show_all
	may show too many or too few files on the io_open_files list.
	This problem can never cause a crash.  io_show_all will remind the
	user of Sherlock when this can occur.
*/

#include "SPP.h"

#include "objlib.h"

#include "SPPio.h"

#include "SPPenv.h"
#include "SPPerr.h"
#include "SPPobj.h"
#include "SPPstr.h"

/*
	Define global variables.
*/
int	io_in_strm_level = -1;
io_strm * io_strm_in_list	= NULL;	/* List of open input streams.*/
io_strm * io_strm_out_list	= NULL;	/* List of open output streams.*/

/*
	Define local variables.
*/
static io_file * io_open_files		= NULL;	/* List of open files.	*/
static bool		 io_closing_flag;

/*
	Function prototypes for local routines.
	Local routines follow all global routines.
*/
static io_file * io_new_file_struct	(io_path * path, FILE * handle);
static io_strm * io_new_strm_struct(io_file * file, int buf_size);
static char *	 io_pr_status	(io_stat status);

/*
	Close a file opened with io_open() or io_create().
*/
void
io_close(io_file * file)
{
	FTAG("io_close");
	
	ASSERT(
		file && file -> io_file_path &&
		file -> io_file_path -> io_path_name);
		
	if (!lst_member(file, &io_open_files)) {
		fatal(ftag);
	}
	
	TRACEPB(ftag, es(file -> io_file_path -> io_path_name); enl());
	
	fclose (file -> io_file_handle);

	/* Free the file structures. */
	if (!lst_remove(file, &io_open_files)) {
		fatal(es("io_close: can't happen"));
	}
	
	obj_free(file->io_file_path->io_path_name);
	obj_free(file -> io_file_path);
	obj_free(file);

	TICKX(ftag);
}

/*
	Close a path.
*/
void
io_close_path (io_path * VAR path)
{
	FTAG("io_close_path");
	STATB(ftag);
	
	/* Closing a already closed path is harmless. */
	if (*path == NULL) {
		return;
	}
	
	ASSERT((*path) -> io_path_name);
	
	obj_free((*path) -> io_path_name);
	obj_free(*path);
	
	/* Mark the path as closed. */
	*path = NULL;
	
	STATX(ftag);
}

/*
	Close a stream and remove it from the appropriate stream list.
*/
void
io_close_strm(io_strm * stream)
{
	FTAG("io_close_strm");

	int n = 0;
		
	TRACEPB(ftag,
		es(stream -> io_strm_file -> io_file_path -> io_path_name);
		enl());
		
	ASSERT(stream);

	
	/* Flush the stream if needed. */
	if (stream -> io_strm_bufc > 0) {
		n = io_write(
			stream -> io_strm_file,
			stream -> io_strm_buffer,
			stream -> io_strm_bufc);
			
		if (n != stream -> io_strm_bufc) {
			error(es("Disk full"));
			return;
		}
	}
	
	/* Close the file. */
	io_close(stream -> io_strm_file);
	
	/* Remove from the stream lists. */
	
	if (
		!lst_remove(stream, &io_strm_in_list) &&
		!lst_remove(stream, &io_strm_out_list)) {
		fatal(es("io_close_strm: can't happen"));
	}
	
	/* Free the file buffer and then the stream itself. */
	obj_free(stream -> io_strm_buffer);
	obj_free(stream);
	
	STATX(ftag);
}

/*
	Open the file for writing only and add to open file list.
	Return an io_file or NULL.
*/
io_file *
io_create(io_path * path, bool cant_flag)
{
	FTAG("io_create");

	FILE *	handle = NULL;
	io_file	* file = NULL;
	bool text_flag = TRUE;		// 3/15/96

	STATB(ftag);
	
	ASSERT(path && path -> io_path_name);

	if (path -> io_path_name[0] == '\0') {
		es("io_create: NULL file name.\n");
		return NULL;
	}
	
	handle = fopen(path -> io_path_name, text_flag ? "w" : "wb");
	if (handle == NULL) {
		if (cant_flag) {
			error(es("Can not create "); es(path -> io_path_name); enl());
		}
		STATX(ftag);
		return NULL;
	}

	// Start of machine-independent part
	
	file = io_new_file_struct(path, handle);
	lst_add(file, &io_open_files);
	
	/* Mark the file open. */
	file -> io_file_stat = IO_OUTPUT;

	STATX(ftag);
	return file;
}

/*
	Open a stream for output.
*/
io_strm *
io_create_strm(io_path * path, bool cant_flag)
{
	FTAG("io_create_strm");

	io_file * file = NULL;
	io_strm * stream = NULL;
	
	file = io_create(path, cant_flag);
	if (file == NULL) {
		return NULL;
	}
	
	stream = io_new_strm_struct(file, IO_BUF_SIZE);
	
	/* Add to list of open output streams. */
	lst_add(stream, &io_strm_out_list);
	
	TRACEP(ftag, eret(); 
		es("stream: "); eptr(stream); ecs();
		es("buffer: "); eptr(stream -> io_strm_buffer); enl());
	
	return stream;
}

/*
	Get the next character from the specified stream.
*/
char
io_getc(io_strm * stream)
{
	FTAG("io_getc");

	reg int n = 0;

	ASSERT(stream);
	
	if (stream -> io_strm_bufc > 0) {
		stream -> io_strm_bufc--;
		return *(stream -> io_strm_bufp)++;
	}

	/* The stream buffer is empty. */
	
	switch(stream -> io_strm_file -> io_file_stat) {

	case IO_EOF:
		/* Continue returning END_FILE until the file is closed. */
		return END_FILE;

	case IO_INPUT:
	
		n = io_read(
			stream -> io_strm_file,
			stream -> io_strm_buffer,
			IO_BUF_SIZE);
			
		if (n > 0) {
			stream -> io_strm_bufc = n - 1;
			stream -> io_strm_bufp = stream -> io_strm_buffer;
			return *(stream -> io_strm_bufp)++;
		}
		else {
			stream -> io_strm_file -> io_file_stat = IO_EOF;
			return END_FILE;
		}

	default:
	
		fatal(
			es("io_getc: Bad file status field: ");
			eint(stream -> io_strm_file -> io_file_stat); enl());

		return END_FILE;
	}
}

/*
	Create a new file structure.
*/
static io_file *
io_new_file_struct(io_path * path, FILE * handle)
{
	FTAG("io_new_file_struct");

	io_file * file = NULL;
	
	STATB(ftag);
	
	file = obj_new(io_file, OBJ_LIB_FILE_STRUCT);
	
	file -> lst_nxt		= NULL;
	file -> io_file_path	= path;
	file -> io_file_handle  = handle;
	file -> io_file_stat	= IO_CLOSED;
	
	TRACEPX(ftag, eret(); io_show_file(file); ecnl());
	return file;
}

/*
	Create a new path structure.
	
	Massage the name if required by the operating system.
*/
io_path *
io_new_path_struct(char * path_name)
{
	FTAG("io_new_path_struct");

	io_path * path_p = NULL;
	char * name_p = NULL;
	int	name_length = 0;
	
	STATB(ftag);
	ASSERT(path_name);

	/* Copy the string to a new area. */
	name_length = strlen(path_name) + 2;
	name_p = obj_alloc(name_length, OBJ_LIB_PATH_NAME+M_PROT);
	strcpy(name_p, path_name);
	
	/* Create the node. */
	path_p = obj_new(io_path, OBJ_LIB_PATH_STRUCT);
	path_p -> io_path_name = name_p;
		
	/* Make the name a valid Macintosh folder name. */
	
#ifdef ENV_DOS
	/*  Make the name a valid DOS directory name. */
	if (path_name[strlen(path_name)-1] != '\\') {
		strcat(path_name, "\\", name_length);
	}
#endif
	
#ifdef NOT_YET
	if (path_name[strlen(path_name)-1] != '\\') {
		strcat(p, "\\", name_length);
	}
#endif

	TRACEPX(ftag, eret(); io_show_path(path_p); ecnl());
	return path_p;
}

/*
	Allocate space for a new file node and file buffer.
	Fill in all other fields to neutral values.
*/
static io_strm *
io_new_strm_struct(io_file * file, int buf_size)
{
	FTAG("io_new_strm_struct");

	io_strm * stream_p = NULL;

	TRACEPB(ftag, es("buf_size: "); eint(buf_size); enl());

	stream_p = obj_new(io_strm, OBJ_LIB_IOSTRM_STRUCT);
	stream_p -> io_strm_buffer =
		obj_alloc(buf_size, OBJ_LIB_STRM_BUFFER+M_PROT+M_NOFILL);
	
	stream_p -> lst_nxt		= NULL;
	stream_p -> io_strm_file	= file;
	stream_p -> io_strm_size	= buf_size;
	stream_p -> io_strm_bufp	= stream_p -> io_strm_buffer;
	stream_p -> io_strm_bufc	= 0;
	
	TRACEPX(ftag, eret(); io_show_strm(stream_p); enl());
	return stream_p;
}

/*
	Open the file for reading only.
	Return a handle (int) or NULL.
*/
io_file *
io_open(io_path * path, bool cant_flag)
{
	FTAG("io_open");

	FILE * handle = NULL;
	io_file * file = NULL;
	bool text_flag = TRUE;		// 3/15/96
	
	ASSERT(path);
	
	if (path -> io_path_name[0] == '\0') {
		if (cant_flag) {
			es("io_open: NULL file name.\n");
		}
		return NULL;
	}
	
	TRACEP("io_open", es(path -> io_path_name); enl());
		
	handle = fopen(path -> io_path_name, text_flag ? "r" : "rb");
	if (handle == NULL) {
	
		if (cant_flag) {
			ecnl(); es("Can not open "); es(path -> io_path_name); enl();
		}

		return NULL;
	}
	
	file = io_new_file_struct(path, handle);
	lst_add(file, &io_open_files);
	
	/* Mark the file open. */
	file -> io_file_stat = IO_INPUT;
	
	return file;
}

/*
	Open a stream for output and add it to the list of open streams.
*/
io_strm *
io_open_strm(io_path * path, bool cant_flag)
{
	FTAG("io_open_strm");

	io_file *	file;
	io_strm *	stream;
	
	TRACEPB(ftag, es(path -> io_path_name); enl());
	
	file = io_open(path, cant_flag);
	if (file == NULL) {
		STATX(ftag);
		return NULL;
	}
	
	stream = io_new_strm_struct(file, IO_BUF_SIZE);
	
	/* Add to list of open streams. */
	lst_add(stream, &io_strm_in_list);
	
	TRACEPX(ftag, eret(); eptr(stream); ecs();
		es("buffer: "); eptr(stream -> io_strm_buffer); enl());
	return stream;
}

/*
	Print the status of a stream.
*/
static char *
io_pr_status(io_stat status)
{
	switch(status) {
	case IO_NULL:	return "IO_NULL";
	case IO_CLOSED:	return "IO_CLOSED";
	case IO_INPUT:	return "IO_INPUT";
	case IO_OUTPUT:	return "IO_OUTPUT";
	case IO_EOF:	return "IO_EOF";
	default:		return "Bad file status";
	}
}

/*
	Put a character to the output file.
	This is the second most called routine after io_next().
*/
void
io_putc(char c)
{
	FTAG("io_putc");

	reg int			n;
	reg io_strm *	stream;
	
	stream = io_strm_out_list;
	ASSERT(stream);
	
	if (c == END_FILE) {
		return;
	}
	
#ifdef ENV_MAC
	if (c == '\n') {
		c = '\r';
	}
#endif

	*(stream -> io_strm_bufp)++ = c;
	
	/* Bug fix:  3/6/91 */
	if (++(stream -> io_strm_bufc) >= stream -> io_strm_size) {
	
		n = io_write(
			stream -> io_strm_file,
			stream -> io_strm_buffer,
			IO_BUF_SIZE);
			
		if (n != IO_BUF_SIZE) {
			error(es("Disk full"));
			return;
		}
		
		stream -> io_strm_bufc = 0;
		stream -> io_strm_bufp = stream -> io_strm_buffer;
	}
}

/*
	Put a string to the output file.
*/
void
io_puts(char * s)
{
	SL_DISABLE();
	
	while(*s) {
		io_putc(*s);
		s++;
	}
}

/*
	Read n bytes from the file described by handle into buffer[].
	Return the number of bytes read.
	
	To do:  change int n to long n.
*/
long 
io_read (io_file * file, char * buffer, long n)
{
	FTAG("io_read");
	STATB(ftag);
	ASSERT(file && buffer && n);
	
	n = fread(buffer, 1, n, file -> io_file_handle);
	
	STATX(ftag);
	return n;
}

/*
	Show all open files.
*/
void
io_show_all(bool warn_flag)
{
	reg io_strm	*	stream;
	reg io_file	*	file;
	
	SL_DISABLE();
	
	ecnls(2);
	
	if (warn_flag) {
		es("Warning:  File list unreliable if i/o in progress.\n");
	}
	
	es("Open input streams are...\n");
	lst_every(io_strm, stream, &io_strm_in_list) {
		io_show_strm(stream);
	}
	
	es("Open output streams are...\n");
	lst_every(io_strm, stream, &io_strm_out_list) {
		io_show_strm(stream);
	}
	
	es("Open files are...\n");
	lst_every(io_file, file, &io_open_files) {
		io_show_file(file);
	}
}


/*
	Print an io_file node.
*/
void
io_show_file(io_file * file)
{
	ASSERT(file);
	
	es("file: "); eptr(file);
	es(", status: "); es(io_pr_status(file -> io_file_stat));
	es(", line: "); elong(file -> io_file_line);
	es(", lastc: "); ehex(file -> io_file_lastc); es("(hex)");
	ecnl();
	io_show_path(file -> io_file_path);
}

/*
	Print an io_path node and a blank line.
*/
void
io_show_path(io_path * path)
{
	SL_DISABLE();
	ASSERT(path);

	es("path: "); es(path -> io_path_name); enl();
}

/*
	Print a io_stream node.
*/
void
io_show_strm(io_strm * stream)
{
	SL_DISABLE();
	ASSERT(stream && stream -> io_strm_file);
	
	es("stream: ");		eptr(stream); enl();
	es(" buffer: ");	eptr(stream -> io_strm_buffer); enl();
	es(" bufsize: ");	eint(stream -> io_strm_size); enl();
	es(" bufp: ");		eptr(stream -> io_strm_bufp); enl();
	es(" bufc: ");		eint(stream -> io_strm_bufc); enl();

	io_show_file(stream -> io_strm_file);
}

/*
	Write n bytes from buffer[] to the file described by handle.
	Return the number of bytes written.
*/
long
io_write(register io_file * file, char * buffer, long n)
{
	FTAG("io_write");
	size_t result = 0;
	STATB(ftag);
	ASSERT(file && buffer);

	result = fwrite(
		buffer,
		1, n,
		file -> io_file_handle);

	STATX(ftag);
	return result;
}

/*
	Allocate a buffer object big enough to hold the whole file.
	Read an entire file into the object, close the file and
	return a pointer to the object.
*/
char *
io_read_all(io_path * path_name)
{
	reg long		n;
	reg long		length;		/* Length of file in bytes		*/
	reg char *		buffer;
	reg io_file *	file;
	
	file = io_open(path_name, TRUE);
	if (file == NULL) {
		return NULL;
	}
	
	length = io_length(file);
	if (length == -1L) {
		return NULL;
	}
	
	buffer = obj_alloc(length + 2, OBJ_LIB_ARG_FILE_BUF);
	if (buffer == NULL) {
		return NULL;
	}
	
	n = io_read(file, buffer, length);
	io_close(file);
	
	if (n != length) {
		return NULL;
	}
	
	/*
		End the buffer with a zero byte.
	*/
	buffer[length] = '\0';
	
	return buffer;
}

/*
	Return the length of an open file or -1L.
*/
long
io_length(io_file * file)
{
	FTAG("io_length");
	long value = -1;
	STATB(ftag);

	// ASSERT(file);

	/* Seek to the end of the file. */
	if (fseek(file -> io_file_handle, 0, SEEK_END) != 0) {
		goto error;
	}

	/* Get the value. */
	value = ftell(file -> io_file_handle);
	if (value == -1) {
		goto error;
	}

	/* Seek to the start of the file. */
	fseek(file -> io_file_handle, 0, SEEK_SET);

error:
done:
	TRACEPX(ftag,
		es(file -> io_file_path -> io_path_name); eblank();
		eret(); elong(value); enl());
	return value;
}
