/*
	SPP: List library.

	source:  SPPlst.c
	started: December 14, 1990
	version: March 16, 1996.
*/

/*
	WARNING:	put Sherlock macros only at the very beginning or
				the very end of the routines in this file.
	
	Sherlock macros act like "software interrupts".
	Any routine using Sherlock macros can generate a call to
	end_quit via the event "call back" routine in init.c.
	
	Sherlock macros must be avoided in "list critical sections,"
	i.e., sections in which the state of a list would be invalid if a
	software interrupt were to occur.
*/

#include "SPP.h"

#include "objlib.h"

#include "SPPerr.h"
#include "SPPlst.h"
#include "SPPobj.h"
#include "SPPstr.h"

/*
	Add a node to the beginning of a list.
*/
void
lst_add_(lst_node * obj, lst_node * VAR list)
{
	TRACEP("lst_add",
		es("node: "); eptr(obj);
		es(" list: "); eptr(list);
		es(" tag: "); es(obj_tag(obj)); enl());

	ASSERT(obj && list);
	obj_check2(obj, OBJ_LIST_BEGIN, OBJ_LIST_END, "lst_add");

	/* Critical section. */
	obj -> lst_nxt = *list;
	*list = obj;
}

void
lst2_add_(lst2_node * obj, lst2_node * VAR list)
{

	TRACEP("lst2_add",
		es("node: "); eptr(obj);
		es(" list: "); eptr(list);
		es(" tag: "); es(obj_tag(obj)); enl());

	ASSERT(obj && list);
	obj_check2(obj, OBJ_LIST2_BEGIN, OBJ_LIST2_END, "lst2_add");
	
	/* Critical section. */
	
	/* Allocate the header node if needed. */
	if (*list == NULL) {
		*list = obj_new(lst2_node, OBJ_LIB_LST2_HEAD);
		(*list) -> lst2_nxt = NULL;
		(*list) -> lst2_back = NULL;
	}
	
	obj -> lst2_nxt = (*list) -> lst2_nxt;
	obj -> lst2_back = (*list);
	(*list) -> lst2_nxt = obj;
}

/*
	Append a node to the end of a list.
*/
void
lst_append_(lst_node * obj, lst_node * VAR list)
{
	lst_node * np;
	
	TRACEP("lst_append",
		es("node: "); eptr(obj);
		es(" list: "); eptr(list);
		es(" tag: "); es(obj_tag(obj)); enl());
			
	ASSERT(obj && list);
	obj_check2(obj, OBJ_LIST_BEGIN, OBJ_LIST_END, "lst_append");

	/* Terminate the list properly. */
	obj -> lst_nxt = NULL;
	
	/* Critical section. */

	/* Special case the first insertion. */
	if ((*list) == NULL) {
		*list = obj;
		return;
	}
	
	/* Scan down the list. */
	np = *list;
	while ((np -> lst_nxt) != NULL) {
		np =  np -> lst_nxt;
	}
	
	/* Point at the new last node. */
	np -> lst_nxt = obj;
}

void
lst2_append_(lst2_node * obj, lst2_node * VAR list)
{
	lst2_node * np;
	
	TRACEP("lst2_append",
		es("node: "); eptr(obj);
		es(" list: "); eptr(list);
		es(" tag: "); es(obj_tag(obj)); enl());
			
	ASSERT(obj && list);
	obj_check2(obj, OBJ_LIST2_BEGIN, OBJ_LIST2_END, "lst2_append");

	/* Terminate the list properly. */
	obj -> lst2_nxt = NULL;
	
	/* Allocate the header node if needed. */
	if (*list == NULL) {
		*list = obj_new(lst2_node, OBJ_LIB_LST2_HEAD);
		(*list) -> lst2_nxt = obj;
		(*list) -> lst2_back = NULL;
		obj 	-> lst2_back = *list;
		return;
	}

	/* Scan down the list. */
	np = *list;
	while ((np -> lst2_nxt) != NULL) {
		np =  np -> lst2_nxt;
	}
	
	/* Point at the new last node. */
	np	-> lst2_nxt = obj;
	obj	-> lst2_back = np;
}

/*
	Free all elements of a list.
*/
void
lst_free_all_(lst_node * VAR list)
{
	reg lst_node * np;
	reg lst_node * np2;
	
	TRACEP("lst_free_all", es("list: "); eptr(list); enl());
	ASSERT(list);

	lst_safe_every(lst_node, np, np2, list) {
		obj_free(np);
	}
	*list = NULL;
}

void
lst2_free_all_(lst2_node * VAR list)
{
	reg lst2_node *	np;
	reg lst2_node *	np2;
	
	TRACEP("lst2_free_all", es("list: "); eptr(list); enl());
	ASSERT(list);
	
	lst2_safe_every(lst2_node, np, np2, list) {
		obj_free(np);
	}
	
	/* Free the header, which is always skipped by lst_safe_every. */
	if (*list != NULL) {
		obj_free(*list);
	}
	*list = NULL;
}

/*
	Return TRUE if the node is a member of the list.
*/
bool
lst_member_(lst_node * obj, lst_node * VAR list)
{
	lst_node * np;
	
	TRACEPB("lst_member",
		es("node: "); eptr(obj);
		es(" list: "); eptr(list);
		es(" tag: "); es(obj_tag(obj)); enl());
			
	ASSERT(obj && list);
	obj_check2(obj, OBJ_LIST_BEGIN, OBJ_LIST_END, "lst_member");

	lst_every(lst_node, np, list) {
		if (np == obj) {
			STATX("lst_member");
			return TRUE;
		}
	}
	STATX("lst_member");
	return FALSE;
}

/*
	Return the next element of the list.
*/
lst_node *
lst_next_(lst_node * VAR list)
{
	SL_DISABLE();
	
	if (list == NULL || (*list) == NULL) {
		return NULL;
	}
	else {
		return (*list) -> lst_nxt;
	}
}

lst2_node *
lst2_next_(lst2_node * VAR list)
{
	SL_DISABLE();
	
	if (list == NULL || (*list) == NULL) {
		return NULL;
	}
	else {
		return (*list) -> lst2_nxt;
	}
}

/*
	Remove a node from a (possibly empty) list.
	Removing a NULL node is not an error.
	
	Return TRUE if the node was in fact removed.
*/
bool
lst_remove_(lst_node * obj, lst_node * VAR list)
{
	lst_node * np1, * np2;
	
	TRACEP("lst_remove",
		es("node: "); eptr(obj);
		es(" list: "); eptr(list);
		es(" tag: "); es(obj_tag(obj)); enl());
		
	ASSERT(obj && list);
	obj_check2(obj, OBJ_LIST_BEGIN, OBJ_LIST_END, "lst_remove");
		
	if (*list == NULL) {
		return FALSE;
	}
	
	/* Special case the start of the list. */
	if (*list == obj) {
		*list = obj -> lst_nxt;
		return TRUE;
	}
	
	for (	np1 = *list, np2 = np1 -> lst_nxt;
			np2 != NULL;
			np1 = np2, np2 = np2 -> lst_nxt
	) {
		if (np2 == obj) {
			np1 -> lst_nxt = np2 -> lst_nxt;
			return TRUE;
		}
	}
	return FALSE;
}

void
lst2_remove_(lst2_node * obj)
{
	reg lst2_node * next, * back;
	
	TRACEP("lst2_remove",
		es("node: "); eptr(obj); es(" tag: "); es(obj_tag(obj)); enl());
		
	ASSERT(obj);
	obj_check2(obj, OBJ_LIST2_BEGIN, OBJ_LIST2_END, "lst2_remove");
	
	next = obj -> lst2_nxt;
	back = obj -> lst2_back;
	
	if (back != NULL) {
		back -> lst2_nxt = next;
	}
	if (next != NULL) {
		next -> lst2_back = back;
	}
}

/*
	Print a report of all the nodes of a list.
	
	The function f is used to print each node.
*/
void
lst_report_ (void (*f) (lst_node *), lst_node ** list, char * name)
{
	lst_node * np;
	
	ecnl();
	es("List: "); es(name); es(" at: "); eptr(list); enl();
	lst_every(lst_node, np, list) {
		(*f) (np);
	}
	ecnls(2);
}
	
void
lst2_report_ (void (*f) (lst2_node *), lst2_node ** list, char * name)
{
	lst2_node * np;
	
	ecnl();
	es("List: "); es(name); es(" at: "); eptr(list); enl();
	lst2_every(lst2_node, np, list) {
		(*f) (np);
	}
	ecnls(2);
}
